@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -27,11 +27,16 @@ static int initialized = 0;
 static int cfg_val[CFG_MAX];
 char *xephyr_screen = "800x600";
 char *xephyr_extra_params = "";
+char *xpra_extra_params = "";
+char *xvfb_screen = "800x600x24";
+char *xvfb_extra_params = "";
 char *netfilter_default = NULL;
 
 int checkcfg(int val) {
 	assert(val < CFG_MAX);
 	int line = 0;
+	FILE *fp = NULL;
+	char *ptr;
 
 	if (!initialized) {
 		// initialize defaults
@@ -39,48 +44,41 @@ int checkcfg(int val) {
 		for (i = 0; i < CFG_MAX; i++)
 			cfg_val[i] = 1; // most of them are enabled by default
 		cfg_val[CFG_RESTRICTED_NETWORK] = 0; // disabled by default
-		cfg_val[CFG_FORCE_NONEWPRIVS] = 0; // disabled by default
-		cfg_val[CFG_PRIVATE_BIN_NO_LOCAL] = 0; // disabled by default
-		
-		// open configuration file
-		char *fname;
-		if (asprintf(&fname, "%s/firejail.config", SYSCONFDIR) == -1)
-			errExit("asprintf");
+		cfg_val[CFG_FORCE_NONEWPRIVS] = 0;
+		cfg_val[CFG_PRIVATE_BIN_NO_LOCAL] = 0;
+		cfg_val[CFG_FIREJAIL_PROMPT] = 0;
+		cfg_val[CFG_FOLLOW_SYMLINK_PRIVATE_BIN] = 0;
+		cfg_val[CFG_DISABLE_MNT] = 0;
+		cfg_val[CFG_ARP_PROBES] = DEFAULT_ARP_PROBES;
+		cfg_val[CFG_XPRA_ATTACH] = 0;
 
-		FILE *fp = fopen(fname, "r");
+		// open configuration file
+		const char *fname = SYSCONFDIR "/firejail.config";
+		fp = fopen(fname, "r");
 		if (!fp) {
-#ifdef HAVE_GLOBALCFG			
+#ifdef HAVE_GLOBALCFG
 			fprintf(stderr, "Error: Firejail configuration file %s not found\n", fname);
 			exit(1);
 #else
 			initialized = 1;
 			return	cfg_val[val];
-#endif	
-		}
-		
-		// if the file exists, it should be owned by root
-		struct stat s;
-		if (stat(fname, &s) == -1)
-			errExit("stat");
-		if (s.st_uid != 0) {
-			fprintf(stderr, "Error: configuration file should be owned by root\n");
-			exit(1);
+#endif
 		}
 
 		// read configuration file
 		char buf[MAX_READ];
 		while (fgets(buf,MAX_READ, fp)) {
 			line++;
-			if (*buf == '#' || *buf == '\n') 
+			if (*buf == '#' || *buf == '\n')
 				continue;
 
-			// parse line				
-			char *ptr = line_remove_spaces(buf);
+			// parse line
+			ptr = line_remove_spaces(buf);
 			if (!ptr)
 				continue;
-			
-			// file transfer	
-			if (strncmp(ptr, "file-transfer ", 14) == 0) {
+
+			// file transfer
+			else if (strncmp(ptr, "file-transfer ", 14) == 0) {
 				if (strcmp(ptr + 14, "yes") == 0)
 					cfg_val[CFG_FILE_TRANSFER] = 1;
 				else if (strcmp(ptr + 14, "no") == 0)
@@ -88,6 +86,15 @@ int checkcfg(int val) {
 				else
 					goto errout;
 			}
+			// join
+			else if (strncmp(ptr, "join ", 5) == 0) {
+				if (strcmp(ptr + 5, "yes") == 0)
+					cfg_val[CFG_JOIN] = 1;
+				else if (strcmp(ptr + 5, "no") == 0)
+					cfg_val[CFG_JOIN] = 0;
+				else
+					goto errout;
+			}
 			// x11
 			else if (strncmp(ptr, "x11 ", 4) == 0) {
 				if (strcmp(ptr + 4, "yes") == 0)
@@ -124,12 +131,39 @@ int checkcfg(int val) {
 				else
 					goto errout;
 			}
+			// prompt
+			else if (strncmp(ptr, "firejail-prompt ", 16) == 0) {
+				if (strcmp(ptr + 16, "yes") == 0)
+					cfg_val[CFG_FIREJAIL_PROMPT] = 1;
+				else if (strcmp(ptr + 16, "no") == 0)
+					cfg_val[CFG_FIREJAIL_PROMPT] = 0;
+				else
+					goto errout;
+			}
+			// follow symlink as user
+			else if (strncmp(ptr, "follow-symlink-as-user ", 23) == 0) {
+				if (strcmp(ptr + 23, "yes") == 0)
+					cfg_val[CFG_FOLLOW_SYMLINK_AS_USER] = 1;
+				else if (strcmp(ptr + 23, "no") == 0)
+					cfg_val[CFG_FOLLOW_SYMLINK_AS_USER] = 0;
+				else
+					goto errout;
+			}
+			// follow symlink in private-bin command
+			else if (strncmp(ptr, "follow-symlink-private-bin ", 27) == 0) {
+				if (strcmp(ptr + 27, "yes") == 0)
+					cfg_val[CFG_FOLLOW_SYMLINK_PRIVATE_BIN] = 1;
+				else if (strcmp(ptr + 27, "no") == 0)
+					cfg_val[CFG_FOLLOW_SYMLINK_PRIVATE_BIN] = 0;
+				else
+					goto errout;
+			}
 			// nonewprivs
 			else if (strncmp(ptr, "force-nonewprivs ", 17) == 0) {
 				if (strcmp(ptr + 17, "yes") == 0)
-					cfg_val[CFG_SECCOMP] = 1;
+					cfg_val[CFG_FORCE_NONEWPRIVS] = 1;
 				else if (strcmp(ptr + 17, "no") == 0)
-					cfg_val[CFG_SECCOMP] = 0;
+					cfg_val[CFG_FORCE_NONEWPRIVS] = 0;
 				else
 					goto errout;
 			}
@@ -177,21 +211,23 @@ int checkcfg(int val) {
 				char *end = strchr(fname, ' ');
 				if (end)
 					*end = '\0';
-				
+
 				// is the file present?
 				struct stat s;
 				if (stat(fname, &s) == -1) {
 					fprintf(stderr, "Error: netfilter-default file %s not available\n", fname);
 					exit(1);
 				}
-				
+
+				if (netfilter_default)
+					goto errout;
 				netfilter_default = strdup(fname);
 				if (!netfilter_default)
 					errExit("strdup");
 				if (arg_debug)
 					printf("netfilter default file %s\n", fname);
 			}
-			
+
 			// Xephyr screen size
 			else if (strncmp(ptr, "xephyr-screen ", 14) == 0) {
 				// expecting two numbers and an x between them
@@ -203,7 +239,7 @@ int checkcfg(int val) {
 				if (asprintf(&xephyr_screen, "%dx%d", n1, n2) == -1)
 					errExit("asprintf");
 			}
-	
+
 			// xephyr window title
 			else if (strncmp(ptr, "xephyr-window-title ", 20) == 0) {
 				if (strcmp(ptr + 20, "yes") == 0)
@@ -213,18 +249,55 @@ int checkcfg(int val) {
 				else
 					goto errout;
 			}
-				
+
 			// Xephyr command extra parameters
-			else if (strncmp(ptr, "xephyr-extra-params ", 19) == 0) {
-				xephyr_extra_params = strdup(ptr + 19);
+			else if (strncmp(ptr, "xephyr-extra-params ", 20) == 0) {
+				if (*xephyr_extra_params != '\0')
+					goto errout;
+				xephyr_extra_params = strdup(ptr + 20);
 				if (!xephyr_extra_params)
 					errExit("strdup");
 			}
-			
+
+			// xpra server extra parameters
+			else if (strncmp(ptr, "xpra-extra-params ", 18) == 0) {
+				if (*xpra_extra_params != '\0')
+					goto errout;
+				xpra_extra_params = strdup(ptr + 18);
+				if (!xpra_extra_params)
+					errExit("strdup");
+			}
+
+			// Xvfb screen size
+	            		else if (strncmp(ptr, "xvfb-screen ", 12) == 0) {
+				// expecting three numbers separated by x's
+				unsigned int n1;
+				unsigned int n2;
+				unsigned int n3;
+				int rv = sscanf(ptr + 12, "%ux%ux%u", &n1, &n2, &n3);
+				if (rv != 3)
+					goto errout;
+				if (asprintf(&xvfb_screen, "%ux%ux%u", n1, n2, n3) == -1)
+					errExit("asprintf");
+			}
+
+			// Xvfb extra parameters
+			else if (strncmp(ptr, "xvfb-extra-params ", 18) == 0) {
+				if (*xvfb_extra_params != '\0')
+					goto errout;
+				xvfb_extra_params = strdup(ptr + 18);
+				if (!xvfb_extra_params)
+					errExit("strdup");
+			}
+
 			// quiet by default
 			else if (strncmp(ptr, "quiet-by-default ", 17) == 0) {
 				if (strcmp(ptr + 17, "yes") == 0)
 					arg_quiet = 1;
+				else if (strcmp(ptr + 17, "no") == 0)
+					arg_quiet = 0;
+				else
+					goto errout;
 			}
 			// remount /proc and /sys
 			else if (strncmp(ptr, "remount-proc-sys ", 17) == 0) {
@@ -251,6 +324,14 @@ int checkcfg(int val) {
 				else
 					goto errout;
 			}
+			else if (strncmp(ptr, "private-lib ", 12) == 0) {
+				if (strcmp(ptr + 12, "yes") == 0)
+					cfg_val[CFG_PRIVATE_LIB] = 1;
+				else if (strcmp(ptr + 12, "no") == 0)
+					cfg_val[CFG_PRIVATE_LIB] = 0;
+				else
+					goto errout;
+			}
 			else if (strncmp(ptr, "chroot-desktop ", 15) == 0) {
 				if (strcmp(ptr + 15, "yes") == 0)
 					cfg_val[CFG_CHROOT_DESKTOP] = 1;
@@ -267,20 +348,47 @@ int checkcfg(int val) {
 				else
 					goto errout;
 			}
+			else if (strncmp(ptr, "disable-mnt ", 12) == 0) {
+				if (strcmp(ptr + 12, "yes") == 0)
+					cfg_val[CFG_DISABLE_MNT] = 1;
+				else if (strcmp(ptr + 12, "no") == 0)
+					cfg_val[CFG_DISABLE_MNT] = 0;
+				else
+					goto errout;
+			}
+			// arp probes
+			else if (strncmp(ptr, "arp-probes ", 11) == 0) {
+				int arp_probes = atoi(ptr + 11);
+				if (arp_probes <= 1 || arp_probes > 30)
+					goto errout;
+				cfg_val[CFG_ARP_PROBES] = arp_probes;
+			}
+			// xpra-attach
+			else if (strncmp(ptr, "xpra-attach ", 12) == 0) {
+				if (strcmp(ptr + 12, "yes") == 0)
+					cfg_val[CFG_XPRA_ATTACH] = 1;
+				else if (strcmp(ptr + 12, "no") == 0)
+					cfg_val[CFG_XPRA_ATTACH] = 0;
+				else
+					goto errout;
+			}
 			else
 				goto errout;
-
+			
 			free(ptr);
 		}
 
 		fclose(fp);
-		free(fname);
 		initialized = 1;
 	}
-	
+
 	return cfg_val[val];
-	
+
 errout:
+	assert(ptr);
+	free(ptr);
+	assert(fp);
+	fclose(fp);
 	fprintf(stderr, "Error: invalid line %d in firejail configuration file\n", line );
 	exit(1);
 }
@@ -304,10 +412,6 @@ void print_compiletime_support(void) {
 #endif
 		);
 
-
-
-
-
 	printf("\t- bind support is %s\n",
 #ifdef HAVE_BIND
 		"enabled"
@@ -340,6 +444,14 @@ void print_compiletime_support(void) {
 #endif
 		);
 
+	printf("\t- git install support is %s\n",
+#ifdef HAVE_GIT_INSTALL
+		"enabled"
+#else
+		"disabled"
+#endif
+		);
+
 	printf("\t- networking support is %s\n",
 #ifdef HAVE_NETWORK
 		"enabled"
@@ -348,7 +460,6 @@ void print_compiletime_support(void) {
 #endif
 		);
 
-
 #ifdef HAVE_NETWORK_RESTRICTED
 	printf("\t- networking features are available only to root user\n");
 #endif
@@ -392,4 +503,5 @@ void print_compiletime_support(void) {
 		"disabled"
 #endif
 		);
+
 }
