@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -20,165 +20,18 @@
 #include "firejail.h"
 #include <sys/mount.h>
 #include <sys/stat.h>
+#include <sys/statvfs.h>
+#include <sys/wait.h>
 #include <linux/limits.h>
 #include <fnmatch.h>
 #include <glob.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <errno.h>
-#include <sys/wait.h>
-
 
 static void fs_rdwr(const char *dir);
 
-static void create_dir_as_root(const char *dir, mode_t mode) {
-	assert(dir);
-	if (arg_debug)
-		printf("Creating %s directory\n", dir);
 
-	if (mkdir(dir, mode) == -1)
-		errExit("mkdir");
-	if (chmod(dir, mode) == -1)
-		errExit("chmod");
-
-	ASSERT_PERMS(dir, 0, 0, mode);
-}
-
-static void create_empty_dir(void) {
-	struct stat s;
-	
-	if (stat(RUN_RO_DIR, &s)) {
-		/* coverity[toctou] */
-		if (mkdir(RUN_RO_DIR, S_IRUSR | S_IXUSR) == -1)
-			errExit("mkdir");
-		if (chmod(RUN_RO_DIR, S_IRUSR | S_IXUSR) == -1)
-			errExit("chmod");
-		ASSERT_PERMS(RUN_RO_DIR, 0, 0, S_IRUSR | S_IXUSR);
-	}
-}
-
-static void create_empty_file(void) {
-	struct stat s;
-
-	if (stat(RUN_RO_FILE, &s)) {
-		/* coverity[toctou] */
-		FILE *fp = fopen(RUN_RO_FILE, "w");
-		if (!fp)
-			errExit("fopen");
-
-		SET_PERMS_STREAM(fp, 0, 0, S_IRUSR);
-		fclose(fp);
-	}
-}
-
-// build /run/firejail directory
-void fs_build_firejail_dir(void) {
-	struct stat s;
-
-	// CentOS 6 doesn't have /run directory
-	if (stat(RUN_FIREJAIL_BASEDIR, &s)) {
-		create_dir_as_root(RUN_FIREJAIL_BASEDIR, 0755);
-	}
-
-	// check /run/firejail directory belongs to root end exit if doesn't!
-	if (stat(RUN_FIREJAIL_DIR, &s) == 0) {
-		if (s.st_uid != 0 || s.st_gid != 0) {
-			fprintf(stderr, "Error: non-root %s directory, exiting...\n", RUN_FIREJAIL_DIR);
-			exit(1);
-		}
-	}
-	else {
-		create_dir_as_root(RUN_FIREJAIL_DIR, 0755);
-	}
-	
-	if (stat(RUN_FIREJAIL_NETWORK_DIR, &s)) {
-		create_dir_as_root(RUN_FIREJAIL_NETWORK_DIR, 0755);
-	}
-	
-	if (stat(RUN_FIREJAIL_BANDWIDTH_DIR, &s)) {
-		create_dir_as_root(RUN_FIREJAIL_BANDWIDTH_DIR, 0755);
-	}
-		
-	if (stat(RUN_FIREJAIL_NAME_DIR, &s)) {
-		create_dir_as_root(RUN_FIREJAIL_NAME_DIR, 0755);
-	}
-	
-	if (stat(RUN_FIREJAIL_X11_DIR, &s)) {
-		create_dir_as_root(RUN_FIREJAIL_X11_DIR, 0755);
-	}
-	
-	if (stat(RUN_FIREJAIL_APPIMAGE_DIR, &s)) {
-		create_dir_as_root(RUN_FIREJAIL_APPIMAGE_DIR, 0755);
-	}
-	
-	create_empty_dir();
-	create_empty_file();
-}
-
-
-// build /run/firejail/mnt directory
-static int tmpfs_mounted = 0;
-#ifdef HAVE_CHROOT		
-static void fs_build_remount_mnt_dir(void) {
-	tmpfs_mounted = 0;
-	fs_build_mnt_dir();
-}
-#endif
-
-void fs_build_mnt_dir(void) {
-	struct stat s;
-	fs_build_firejail_dir();
-	
-	// create /run/firejail/mnt directory
-	if (stat(RUN_MNT_DIR, &s)) {
-		create_dir_as_root(RUN_MNT_DIR, 0755);
-	}
-
-	// ... and mount tmpfs on top of it
-	if (!tmpfs_mounted) {
-		// mount tmpfs on top of /run/firejail/mnt
-		if (arg_debug)
-			printf("Mounting tmpfs on %s directory\n", RUN_MNT_DIR);
-		if (mount("tmpfs", RUN_MNT_DIR, "tmpfs", MS_NOSUID | MS_STRICTATIME | MS_REC,  "mode=755,gid=0") < 0)
-			errExit("mounting /run/firejail/mnt");
-		tmpfs_mounted = 1;
-		fs_logger2("tmpfs", RUN_MNT_DIR);
-	}
-}
-
-// grab a copy of cp command
-void fs_build_cp_command(void) {
-	struct stat s;
-	fs_build_mnt_dir();
-	if (stat(RUN_CP_COMMAND, &s)) {
-		char* fname = realpath("/bin/cp", NULL);
-		if (fname == NULL) {
-			fprintf(stderr, "Error: /bin/cp not found\n");
-			exit(1);
-		}
-		if (stat(fname, &s)) {
-			fprintf(stderr, "Error: /bin/cp not found\n");
-			exit(1);
-		}
-		if (is_link(fname)) {
-			fprintf(stderr, "Error: invalid /bin/cp file\n");
-			exit(1);
-		}
-		int rv = copy_file(fname, RUN_CP_COMMAND, 0, 0, 0755); // root needed
-		if (rv) {
-			fprintf(stderr, "Error: cannot access /bin/cp\n");
-			exit(1);
-		}
-		ASSERT_PERMS(RUN_CP_COMMAND, 0, 0, 0755);
-			
-		free(fname);
-	}
-}
-
-// delete the temporary cp command
-void fs_delete_cp_command(void) {
-	unlink(RUN_CP_COMMAND);
-}
 
 //***********************************************
 // process profile file
@@ -203,10 +56,7 @@ static void disable_file(OPERATION op, const char *filename) {
 	assert(filename);
 	assert(op <OPERATION_MAX);
 	last_disable = UNSUCCESSFUL;
-	
-	// rebuild /run/firejail directory in case tmpfs was mounted on top of /run
-	fs_build_firejail_dir();
-	
+
 	// Resolve all symlinks
 	char* fname = realpath(filename, NULL);
 	if (fname == NULL && errno != EACCES) {
@@ -238,17 +88,17 @@ static void disable_file(OPERATION op, const char *filename) {
 			if (arg_debug)
 				printf("Warning (blacklisting): %s is an invalid file, skipping...\n", filename);
 		}
-				
+
 		return;
 	}
-	
+
 	// if the file is not present, do nothing
 	struct stat s;
 	if (fname == NULL)
 		return;
 	if (stat(fname, &s) == -1) {
 		if (arg_debug)
-			printf("Warning: %s does not exist, skipping...\n", fname);
+			fwarning("%s does not exist, skipping...\n", fname);
 		free(fname);
 		return;
 	}
@@ -258,12 +108,16 @@ static void disable_file(OPERATION op, const char *filename) {
 		// some distros put all executables under /usr/bin and make /bin a symbolic link
 		if ((strcmp(fname, "/bin") == 0 || strcmp(fname, "/usr/bin") == 0) &&
 		      is_link(filename) &&
-		      S_ISDIR(s.st_mode))
-			fprintf(stderr, "Warning: %s directory link was not blacklisted\n", filename);
-			
+		      S_ISDIR(s.st_mode)) {
+			fwarning("%s directory link was not blacklisted\n", filename);
+		}
 		else {
-			if (arg_debug)
-				printf("Disable %s\n", fname);
+			if (arg_debug) {
+				if (strcmp(filename, fname))
+					printf("Disable %s (requested %s)\n", fname, filename);
+				else
+					printf("Disable %s\n", fname);
+			}
 			else if (arg_debug_blacklists) {
 				printf("Disable %s", fname);
 				if (op == BLACKLIST_FILE)
@@ -271,6 +125,7 @@ static void disable_file(OPERATION op, const char *filename) {
 				else
 					printf(" - no logging\n");
 			}
+
 			if (S_ISDIR(s.st_mode)) {
 				if (mount(RUN_RO_DIR, fname, "none", MS_BIND, "mode=400,gid=0") < 0)
 					errExit("disable file");
@@ -313,12 +168,14 @@ static void disable_file(OPERATION op, const char *filename) {
 				errExit("mounting tmpfs");
 			/* coverity[toctou] */
 			if (chown(fname, s.st_uid, s.st_gid) == -1)
+				errExit("mounting tmpfs chown");
+			if (chmod(fname, s.st_mode) == -1)
 				errExit("mounting tmpfs chmod");
 			last_disable = SUCCESSFUL;
 			fs_logger2("tmpfs", fname);
 		}
 		else
-			printf("Warning: %s is not a directory; cannot mount a tmpfs on top of it.\n", fname);
+			fwarning("%s is not a directory; cannot mount a tmpfs on top of it.\n", fname);
 	}
 	else
 		assert(0);
@@ -362,6 +219,15 @@ static void globbing(OPERATION op, const char *pattern, const char *noblacklist[
 				exit(1);
 			}
 		}
+
+		// We don't usually need to blacklist things in private home directories
+		if (okay_to_blacklist
+		 && cfg.homedir
+		 && arg_private
+		 && (!arg_allow_private_blacklist)
+		 && (strncmp(path, cfg.homedir, strlen(cfg.homedir)) == 0))
+			okay_to_blacklist = false;
+
 		if (okay_to_blacklist)
 			disable_file(op, path);
 		else if (arg_debug)
@@ -371,14 +237,14 @@ static void globbing(OPERATION op, const char *pattern, const char *noblacklist[
 }
 
 
-// blacklist files or directoies by mounting empty files on top of them
+// blacklist files or directories by mounting empty files on top of them
 void fs_blacklist(void) {
 	char *homedir = cfg.homedir;
 	assert(homedir);
 	ProfileEntry *entry = cfg.profile;
 	if (!entry)
 		return;
-		
+
 	size_t noblacklist_c = 0;
 	size_t noblacklist_m = 32;
 	char **noblacklist = calloc(noblacklist_m, sizeof(*noblacklist));
@@ -391,32 +257,26 @@ void fs_blacklist(void) {
 		char *ptr;
 
 		// whitelist commands handled by fs_whitelist()
-		if (strncmp(entry->data, "whitelist ", 10) == 0 || *entry->data == '\0') {
+		if (strncmp(entry->data, "whitelist ", 10) == 0 ||
+		    strncmp(entry->data, "nowhitelist ", 12) == 0 ||
+		   *entry->data == '\0') {
 			entry = entry->next;
 			continue;
 		}
 
 		// process bind command
 		if (strncmp(entry->data, "bind ", 5) == 0)  {
+			struct stat s;
 			char *dname1 = entry->data + 5;
 			char *dname2 = split_comma(dname1);
-			if (dname2 == NULL) {
-				fprintf(stderr, "Error: second directory missing in bind command\n");
+			if (dname2 == NULL ||
+			    stat(dname1, &s) == -1 ||
+			    stat(dname2, &s) == -1) {
+				fprintf(stderr, "Error: invalid bind command, directory missing\n");
 				entry = entry->next;
 				continue;
 			}
-			struct stat s;
-			if (stat(dname1, &s) == -1) {
-				fprintf(stderr, "Error: cannot find %s for bind command\n", dname1);
-				entry = entry->next;
-				continue;
-			}
-			if (stat(dname2, &s) == -1) {
-				fprintf(stderr, "Error: cannot find %s for bind command\n", dname2);
-				entry = entry->next;
-				continue;
-			}
-			
+
 			// mount --bind olddir newdir
 			if (arg_debug)
 				printf("Mount-bind %s on top of %s\n", dname1, dname2);
@@ -424,38 +284,44 @@ void fs_blacklist(void) {
 			if (mount(dname1, dname2, NULL, MS_BIND|MS_REC, NULL) < 0)
 				errExit("mount bind");
 			/* coverity[toctou] */
-			if (chown(dname2, s.st_uid, s.st_gid) == -1)
-				errExit("mount-bind chown");
-			/* coverity[toctou] */
-			if (chmod(dname2, s.st_mode) == -1)
-				errExit("mount-bind chmod");
-				
+			if (set_perms(dname2,  s.st_uid, s.st_gid,s.st_mode))
+				errExit("set_perms");
+
 			entry = entry->next;
 			continue;
 		}
 
 		// Process noblacklist command
 		if (strncmp(entry->data, "noblacklist ", 12) == 0) {
-			char **paths = build_paths();
-
-			char *enames[sizeof(paths)+1] = {0};
-			int i = 0;
+			char **enames;
+			int i;
 
 			if (strncmp(entry->data + 12, "${PATH}", 7) == 0) {
 				// expand ${PATH} macro
-				while (paths[i] != NULL) {
-					if (asprintf(&enames[i], "%s%s", paths[i], entry->data + 19) == -1)
+				char **paths = build_paths();
+				unsigned int npaths = count_paths();
+				enames = calloc(npaths, sizeof(char *));
+				if (!enames)
+					errExit("calloc");
+
+				for (i = 0; paths[i]; i++) {
+					if (asprintf(&enames[i], "%s%s", paths[i],
+						entry->data + 19) == -1)
 						errExit("asprintf");
-					i++;
 				}
-			} else {
+				assert(enames[npaths-1] == 0);
+
+			}
+			else {
 				// expand ${HOME} macro if found or pass as is
+				enames = calloc(2, sizeof(char *));
+				if (!enames)
+					errExit("calloc");
 				enames[0] = expand_home(entry->data + 12, homedir);
-				enames[1] = NULL;
+				assert(enames[1] == 0);
 			}
 
-			i = 0;
-			while (enames[i] != NULL) {
+			for (i = 0; enames[i]; i++) {
 				if (noblacklist_c >= noblacklist_m) {
 					noblacklist_m *= 2;
 					noblacklist = realloc(noblacklist, sizeof(*noblacklist) * noblacklist_m);
@@ -463,12 +329,9 @@ void fs_blacklist(void) {
 						errExit("failed increasing memory for noblacklist entries");
 				}
 				noblacklist[noblacklist_c++] = enames[i];
-				i++;
 			}
 
-			while (enames[i] != NULL) {
-				free(enames[i]);
-			}
+			free(enames);
 
 			entry = entry->next;
 			continue;
@@ -486,33 +349,33 @@ void fs_blacklist(void) {
 		else if (strncmp(entry->data, "read-only ", 10) == 0) {
 			ptr = entry->data + 10;
 			op = MOUNT_READONLY;
-		}			
+		}
 		else if (strncmp(entry->data, "read-write ", 11) == 0) {
 			ptr = entry->data + 11;
 			op = MOUNT_RDWR;
-		}			
+		}
 		else if (strncmp(entry->data, "noexec ", 7) == 0) {
 			ptr = entry->data + 7;
 			op = MOUNT_NOEXEC;
-		}			
+		}
 		else if (strncmp(entry->data, "tmpfs ", 6) == 0) {
 			ptr = entry->data + 6;
 			op = MOUNT_TMPFS;
-		}			
+		}
 		else if (strncmp(entry->data, "mkdir ", 6) == 0) {
 			EUID_USER();
 			fs_mkdir(entry->data + 6);
 			EUID_ROOT();
 			entry = entry->next;
 			continue;
-		}			
+		}
 		else if (strncmp(entry->data, "mkfile ", 7) == 0) {
 			EUID_USER();
 			fs_mkfile(entry->data + 7);
 			EUID_ROOT();
 			entry = entry->next;
 			continue;
-		}			
+		}
 		else {
 			fprintf(stderr, "Error: invalid profile line %s\n", entry->data);
 			entry = entry->next;
@@ -552,6 +415,15 @@ void fs_blacklist(void) {
         free(noblacklist);
 }
 
+static int get_mount_flags(const char *path, unsigned long *flags) {
+	struct statvfs buf;
+
+	if (statvfs(path, &buf) < 0)
+		return -errno;
+	*flags = buf.f_flag;
+	return 0;
+}
+
 //***********************************************
 // mount namespace
 //***********************************************
@@ -563,11 +435,15 @@ void fs_rdonly(const char *dir) {
 	struct stat s;
 	int rv = stat(dir, &s);
 	if (rv == 0) {
+		unsigned long flags = 0;
+		get_mount_flags(dir, &flags);
+		if ((flags & MS_RDONLY) == MS_RDONLY)
+			return;
+		flags |= MS_RDONLY;
 		// mount --bind /bin /bin
-		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0)
-			errExit("mount read-only");
 		// mount --bind -o remount,ro /bin
-		if (mount(NULL, dir, NULL, MS_BIND|MS_REMOUNT|MS_RDONLY|MS_REC, NULL) < 0)
+		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0 ||
+		    mount(NULL, dir, NULL, flags|MS_BIND|MS_REMOUNT|MS_REC, NULL) < 0)
 			errExit("mount read-only");
 		fs_logger2("read-only", dir);
 	}
@@ -582,15 +458,19 @@ static void fs_rdwr(const char *dir) {
 		// if the file is outside /home directory, allow only root user
 		uid_t u = getuid();
 		if (u != 0 && s.st_uid != u) {
-			fprintf(stderr, "Warning: you are not allowed to change %s to read-write\n", dir);
+			fwarning("you are not allowed to change %s to read-write\n", dir);
 			return;
 		}
-		
+
 		// mount --bind /bin /bin
-		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0)
-			errExit("mount read-write");
 		// mount --bind -o remount,rw /bin
-		if (mount(NULL, dir, NULL, MS_BIND|MS_REMOUNT|MS_REC, NULL) < 0)
+		unsigned long flags = 0;
+		get_mount_flags(dir, &flags);
+		if ((flags & MS_RDONLY) == 0)
+			return;
+		flags &= ~MS_RDONLY;
+		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0 ||
+		    mount(NULL, dir, NULL, flags|MS_BIND|MS_REMOUNT|MS_REC, NULL) < 0)
 			errExit("mount read-write");
 		fs_logger2("read-write", dir);
 	}
@@ -603,37 +483,28 @@ void fs_noexec(const char *dir) {
 	int rv = stat(dir, &s);
 	if (rv == 0) {
 		// mount --bind /bin /bin
-		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0)
-			errExit("mount noexec");
 		// mount --bind -o remount,ro /bin
-		if (mount(NULL, dir, NULL, MS_BIND|MS_REMOUNT|MS_NOEXEC|MS_NODEV|MS_NOSUID|MS_REC, NULL) < 0)
-			errExit("mount read-only");
+		unsigned long flags = 0;
+		get_mount_flags(dir, &flags);
+		if ((flags & (MS_NOEXEC|MS_NODEV|MS_NOSUID)) == (MS_NOEXEC|MS_NODEV|MS_NOSUID))
+			return;
+		flags |= MS_NOEXEC|MS_NODEV|MS_NOSUID;
+		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0 ||
+		    mount(NULL, dir, NULL, flags|MS_BIND|MS_REMOUNT|MS_REC, NULL) < 0)
+			errExit("mount noexec");
 		fs_logger2("noexec", dir);
 	}
 }
 
-
-
-void fs_rdonly_noexit(const char *dir) {
-	assert(dir);
-	// check directory exists
-	struct stat s;
-	int rv = stat(dir, &s);
-	if (rv == 0) {
-		int merr = 0;
-		// mount --bind /bin /bin
-		if (mount(dir, dir, NULL, MS_BIND|MS_REC, NULL) < 0)
-			merr = 1;
-		// mount --bind -o remount,ro /bin
-		if (mount(NULL, dir, NULL, MS_BIND|MS_REMOUNT|MS_RDONLY|MS_REC, NULL) < 0)
-			merr = 1;
-		if (merr)
-			fprintf(stderr, "Warning: cannot mount %s read-only\n", dir); 
-		else
-			fs_logger2("read-only", dir);
-	}
+// Disable /mnt, /media, /run/mount and /run/media access
+void fs_mnt(void) {
+	disable_file(BLACKLIST_FILE, "/mnt");
+	disable_file(BLACKLIST_FILE, "/media");
+	disable_file(BLACKLIST_FILE, "/run/mount");
+	disable_file(BLACKLIST_FILE, "//run/media");
 }
 
+
 // mount /proc and /sys directories
 void fs_proc_sys_dev_boot(void) {
 	if (arg_debug)
@@ -643,10 +514,8 @@ void fs_proc_sys_dev_boot(void) {
 	fs_logger("remount /proc");
 
 	// remount /proc/sys readonly
-	if (mount("/proc/sys", "/proc/sys", NULL, MS_BIND | MS_REC, NULL) < 0)
-		errExit("mounting /proc/sys");
-
-	if (mount(NULL, "/proc/sys", NULL, MS_BIND | MS_REMOUNT | MS_RDONLY | MS_REC, NULL) < 0)
+	if (mount("/proc/sys", "/proc/sys", NULL, MS_BIND | MS_REC, NULL) < 0 ||
+	    mount(NULL, "/proc/sys", NULL, MS_BIND | MS_REMOUNT | MS_RDONLY | MS_NOSUID | MS_NOEXEC | MS_NODEV | MS_REC, NULL) < 0)
 		errExit("mounting /proc/sys");
 	fs_logger("read-only /proc/sys");
 
@@ -655,18 +524,16 @@ void fs_proc_sys_dev_boot(void) {
 	if (arg_debug)
 		printf("Remounting /sys directory\n");
 	if (umount2("/sys", MNT_DETACH) < 0)
-		fprintf(stderr, "Warning: failed to unmount /sys\n");
-	else {
-		if (mount("sysfs", "/sys", "sysfs", MS_RDONLY|MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_REC, NULL) < 0)
-			fprintf(stderr, "Warning: failed to mount /sys\n");
-		else
-			fs_logger("remount /sys");
-	}
-		
+		fwarning("failed to unmount /sys\n");
+	if (mount("sysfs", "/sys", "sysfs", MS_RDONLY|MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_REC, NULL) < 0)
+		fwarning("failed to mount /sys\n");
+	else
+		fs_logger("remount /sys");
+
 	disable_file(BLACKLIST_FILE, "/sys/firmware");
 	disable_file(BLACKLIST_FILE, "/sys/hypervisor");
 	{ // allow user access to /sys/fs if "--noblacklist=/sys/fs" is present on the command line
-		EUID_USER();	
+		EUID_USER();
 		profile_add("blacklist /sys/fs");
 		EUID_ROOT();
 	}
@@ -677,11 +544,11 @@ void fs_proc_sys_dev_boot(void) {
 	disable_file(BLACKLIST_FILE, "/sys/kernel/uevent_helper");
 
 	// various /proc/sys files
-	disable_file(BLACKLIST_FILE, "/proc/sys/security");	
-	disable_file(BLACKLIST_FILE, "/proc/sys/efi/vars");	
-	disable_file(BLACKLIST_FILE, "/proc/sys/fs/binfmt_misc");	
-	disable_file(BLACKLIST_FILE, "/proc/sys/kernel/core_pattern");	
-	disable_file(BLACKLIST_FILE, "/proc/sys/kernel/modprobe");	
+	disable_file(BLACKLIST_FILE, "/proc/sys/security");
+	disable_file(BLACKLIST_FILE, "/proc/sys/efi/vars");
+	disable_file(BLACKLIST_FILE, "/proc/sys/fs/binfmt_misc");
+	disable_file(BLACKLIST_FILE, "/proc/sys/kernel/core_pattern");
+	disable_file(BLACKLIST_FILE, "/proc/sys/kernel/modprobe");
 	disable_file(BLACKLIST_FILE, "/proc/sysrq-trigger");
 	disable_file(BLACKLIST_FILE, "/proc/sys/kernel/hotplug");
 	disable_file(BLACKLIST_FILE, "/proc/sys/vm/panic_on_oom");
@@ -689,15 +556,15 @@ void fs_proc_sys_dev_boot(void) {
 	// various /proc files
 	disable_file(BLACKLIST_FILE, "/proc/irq");
 	disable_file(BLACKLIST_FILE, "/proc/bus");
-	disable_file(BLACKLIST_FILE, "/proc/config.gz");	
-	disable_file(BLACKLIST_FILE, "/proc/sched_debug");	
-	disable_file(BLACKLIST_FILE, "/proc/timer_list");	
-	disable_file(BLACKLIST_FILE, "/proc/timer_stats");	
+	disable_file(BLACKLIST_FILE, "/proc/config.gz");
+	disable_file(BLACKLIST_FILE, "/proc/sched_debug");
+	disable_file(BLACKLIST_FILE, "/proc/timer_list");
+	disable_file(BLACKLIST_FILE, "/proc/timer_stats");
 	disable_file(BLACKLIST_FILE, "/proc/kcore");
 	disable_file(BLACKLIST_FILE, "/proc/kallsyms");
 	disable_file(BLACKLIST_FILE, "/proc/mem");
 	disable_file(BLACKLIST_FILE, "/proc/kmem");
-	
+
 	// remove kernel symbol information
 	if (!arg_allow_debuggers) {
 		disable_file(BLACKLIST_FILE, "/usr/src/linux");
@@ -705,13 +572,44 @@ void fs_proc_sys_dev_boot(void) {
 		disable_file(BLACKLIST_FILE, "/usr/lib/debug");
 		disable_file(BLACKLIST_FILE, "/boot");
 	}
-		
+
 	// disable /selinux
 	disable_file(BLACKLIST_FILE, "/selinux");
-	
+
 	// disable /dev/port
 	disable_file(BLACKLIST_FILE, "/dev/port");
-	
+
+
+
+	// disable various ipc sockets in /run/user
+	struct stat s;
+
+	char *fname;
+	if (asprintf(&fname, "/run/user/%d", getuid()) == -1)
+		errExit("asprintf");
+	if (is_dir(fname)) { // older distros don't have this directory
+		// disable /run/user/{uid}/gnupg
+		char *fnamegpg;
+		if (asprintf(&fnamegpg, "/run/user/%d/gnupg", getuid()) == -1)
+			errExit("asprintf");
+		if (stat(fnamegpg, &s) == -1)
+		    mkdir_attr(fnamegpg, 0700, getuid(), getgid());
+		if (stat(fnamegpg, &s) == 0)
+			disable_file(BLACKLIST_FILE, fnamegpg);
+		free(fnamegpg);
+
+		// disable /run/user/{uid}/systemd
+		char *fnamesysd;
+		if (asprintf(&fnamesysd, "/run/user/%d/systemd", getuid()) == -1)
+			errExit("asprintf");
+		if (stat(fnamesysd, &s) == -1)
+			mkdir_attr(fnamesysd, 0755, getuid(), getgid());
+		if (stat(fnamesysd, &s) == 0)
+			disable_file(BLACKLIST_FILE, fnamesysd);
+		free(fnamesysd);
+	}
+	free(fname);
+
 	if (getuid() != 0) {
 		// disable /dev/kmsg and /proc/kmsg
 		disable_file(BLACKLIST_FILE, "/dev/kmsg");
@@ -729,7 +627,7 @@ static void disable_config(void) {
 	if (stat(fname, &s) == 0)
 		disable_file(BLACKLIST_FILE, fname);
 	free(fname);
-	
+
 	// disable run time information
 	if (stat(RUN_FIREJAIL_NETWORK_DIR, &s) == 0)
 		disable_file(BLACKLIST_FILE, RUN_FIREJAIL_NETWORK_DIR);
@@ -745,7 +643,7 @@ static void disable_config(void) {
 // build a basic read-only filesystem
 void fs_basic_fs(void) {
 	uid_t uid = getuid();
-	
+
 	if (arg_debug)
 		printf("Mounting read-only /bin, /sbin, /lib, /lib32, /lib64, /usr");
 	if (!arg_writable_etc) {
@@ -770,19 +668,21 @@ void fs_basic_fs(void) {
 	fs_rdonly("/usr");
 
 	// update /var directory in order to support multiple sandboxes running on the same root directory
-	if (!arg_private_dev)
-		fs_dev_shm();
 	fs_var_lock();
 	fs_var_tmp();
-	fs_var_log();
+	if (!arg_writable_var_log)
+		fs_var_log();
+	else
+		fs_rdwr("/var/log");
+
 	fs_var_lib();
 	fs_var_cache();
 	fs_var_utmp();
 	fs_machineid();
-	
+
 	// don't leak user information
 	restrict_users();
-	
+
 	// when starting as root, firejail config is not disabled;
 	// this mode could be used to install and test new software by chaining
 	// firejail sandboxes (firejail --force)
@@ -800,7 +700,7 @@ char *fs_check_overlay_dir(const char *subdirname, int allow_reuse) {
 	// create ~/.firejail directory
 	if (asprintf(&dirname, "%s/.firejail", cfg.homedir) == -1)
 		errExit("asprintf");
-		
+
 	if (is_link(dirname)) {
 		fprintf(stderr, "Error: invalid ~/.firejail directory\n");
 		exit(1);
@@ -813,7 +713,7 @@ char *fs_check_overlay_dir(const char *subdirname, int allow_reuse) {
 		if (child == 0) {
 			// drop privileges
 			drop_privs(0);
-	
+
 			// create directory
 			if (mkdir(dirname, 0700))
 				errExit("mkdir");
@@ -895,7 +795,7 @@ void fs_overlayfs(void) {
 		fprintf(stderr, "Error: cannot extract Linux kernel version: %s\n", u.version);
 		exit(1);
 	}
-	
+
 	if (arg_debug)
 		printf("Linux kernel version %d.%d\n", major, minor);
 	int oldkernel = 0;
@@ -905,18 +805,11 @@ void fs_overlayfs(void) {
 	}
 	if (major == 3 && minor < 18)
 		oldkernel = 1;
-	
-	// build overlay directories
-	fs_build_mnt_dir();
 
 	char *oroot;
 	if(asprintf(&oroot, "%s/oroot", RUN_MNT_DIR) == -1)
 		errExit("asprintf");
-	if (mkdir(oroot, 0755))
-		errExit("mkdir");
-	if (chmod(oroot, 0755) == -1)
-		errExit("chmod");
-	ASSERT_PERMS(oroot, 0, 0, 0755);
+	mkdir_attr(oroot, 0755, 0, 0);
 
 	struct stat s;
 	char *basedir = RUN_MNT_DIR;
@@ -932,6 +825,7 @@ void fs_overlayfs(void) {
 			}
 		}
 		else {
+			/* coverity[toctou] */
 			if (mkdir(basedir, 0755) != 0) {
 				fprintf(stderr, "Error: cannot create overlay directory\n");
 				exit(1);
@@ -945,30 +839,22 @@ void fs_overlayfs(void) {
 
 	// no need to check arg_overlay_reuse
 	if (stat(odiff, &s) != 0) {
-		if (mkdir(odiff, 0755))
-			errExit("mkdir");
+		mkdir_attr(odiff, 0755, 0, 0);
 	}
+	else if (set_perms(odiff, 0, 0, 0755))
+		errExit("set_perms");
 
-	if (chown(odiff, 0, 0) < 0)
-		errExit("chown");
-	if (chmod(odiff, 0755) < 0)
-		errExit("chmod");
-	
 	char *owork;
 	if(asprintf(&owork, "%s/owork", basedir) == -1)
 		errExit("asprintf");
 
 	// no need to check arg_overlay_reuse
 	if (stat(owork, &s) != 0) {
-		if (mkdir(owork, 0755))
-			errExit("mkdir");
+		mkdir_attr(owork, 0755, 0, 0);
 	}
-
-	if (chown(owork, 0, 0) < 0)
+	else if (set_perms(owork, 0, 0, 0755))
 		errExit("chown");
-	if (chmod(owork, 0755) < 0)
-		errExit("chmod");
-	
+
 	// mount overlayfs
 	if (arg_debug)
 		printf("Mounting OverlayFS\n");
@@ -988,11 +874,11 @@ void fs_overlayfs(void) {
 			errExit("asprintf");
 		if (mount("overlay", oroot, "overlay", MS_MGC_VAL, option) < 0)
 			errExit("mounting overlayfs");
-			
+
 		//***************************
 		// issue #263 start code
 		// My setup has a separate mount point for /home. When the overlay is mounted,
-		// the overlay does not contain the original /home contents. 
+		// the overlay does not contain the original /home contents.
 		// I added code to create a second overlay for /home if the overlay home dir is empty and this seems to work
 		// @dshmgh, Jan 2016
 		{
@@ -1001,56 +887,48 @@ void fs_overlayfs(void) {
 			char *hroot;
 			char *hdiff;
 			char *hwork;
-		
+
 			// dons add debug
 			if (arg_debug) printf ("DEBUG: chroot dirs are oroot %s  odiff %s  owork %s\n",oroot,odiff,owork);
-		
+
 			// BEFORE NEXT, WE NEED TO TEST IF /home has any contents or do we need to mount it?
 			// must create var for oroot/cfg.homedir
 			if (asprintf(&overlayhome,"%s%s",oroot,cfg.homedir) == -1)
 				errExit("asprintf");
 			if (arg_debug) printf ("DEBUG: overlayhome var holds ##%s##\n",overlayhome);
-		
+
 			// if no homedir in overlay -- create another overlay for /home
 			if (stat(overlayhome, &s) == -1) {
-		
+
 				if(asprintf(&hroot, "%s/oroot/home", RUN_MNT_DIR) == -1)
 					errExit("asprintf");
-		
+
 				if(asprintf(&hdiff, "%s/hdiff", basedir) == -1)
 					errExit("asprintf");
 
 				// no need to check arg_overlay_reuse
 				if (stat(hdiff, &s) != 0) {
-					if (mkdir(hdiff, S_IRWXU | S_IRWXG | S_IRWXO))
-						errExit("mkdir");
+					mkdir_attr(hdiff, S_IRWXU | S_IRWXG | S_IRWXO, 0, 0);
 				}
+				else if (set_perms(hdiff, 0, 0, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH))
+					errExit("set_perms");
 
-				if (chown(hdiff, 0, 0) < 0)
-					errExit("chown");
-				if (chmod(hdiff, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
-					errExit("chmod");
-		
 				if(asprintf(&hwork, "%s/hwork", basedir) == -1)
 					errExit("asprintf");
 
 				// no need to check arg_overlay_reuse
 				if (stat(hwork, &s) != 0) {
-					if (mkdir(hwork, S_IRWXU | S_IRWXG | S_IRWXO))
-						errExit("mkdir");
+					mkdir_attr(hwork, S_IRWXU | S_IRWXG | S_IRWXO, 0, 0);
 				}
+				else if (set_perms(hwork, 0, 0, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH))
+					errExit("set_perms");
 
-				if (chown(hwork, 0, 0) < 0)
-					errExit("chown");
-				if (chmod(hwork, S_IRWXU  | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) < 0)
-					errExit("chmod");
-		
 				// no homedir in overlay so now mount another overlay for /home
 				if (asprintf(&option, "lowerdir=/home,upperdir=%s,workdir=%s", hdiff, hwork) == -1)
 					errExit("asprintf");
 				if (mount("overlay", hroot, "overlay", MS_MGC_VAL, option) < 0)
 					errExit("mounting overlayfs for mounted home directory");
-		
+
 				printf("OverlayFS for /home configured in %s directory\n", basedir);
 			} // stat(overlayhome)
 			free(overlayhome);
@@ -1058,8 +936,9 @@ void fs_overlayfs(void) {
 		// issue #263 end code
 		//***************************
 	}
-	printf("OverlayFS configured in %s directory\n", basedir);
-	
+	if (!arg_quiet)
+		printf("OverlayFS configured in %s directory\n", basedir);
+
 	// mount-bind dev directory
 	if (arg_debug)
 		printf("Mounting /dev\n");
@@ -1088,22 +967,29 @@ void fs_overlayfs(void) {
 		if (asprintf(&x11, "%s/tmp/.X11-unix", oroot) == -1)
 			errExit("asprintf");
 		if (mount("/tmp/.X11-unix", x11, NULL, MS_BIND|MS_REC, NULL) < 0)
-			fprintf(stderr, "Warning: cannot mount /tmp/.X11-unix in overlay\n");
+			fwarning("cannot mount /tmp/.X11-unix in overlay\n");
 		else
 			fs_logger("whitelist /tmp/.X11-unix");
 		free(x11);
 	}
 
 	// chroot in the new filesystem
+#ifdef HAVE_GCOV
+	__gcov_flush();
+#endif
 	if (chroot(oroot) == -1)
 		errExit("chroot");
 
 	// update /var directory in order to support multiple sandboxes running on the same root directory
-	if (!arg_private_dev)
-		fs_dev_shm();
+//	if (!arg_private_dev)
+//		fs_dev_shm();
 	fs_var_lock();
 	fs_var_tmp();
-	fs_var_log();
+	if (!arg_writable_var_log)
+		fs_var_log();
+	else
+		fs_rdwr("/var/log");
+
 	fs_var_lib();
 	fs_var_cache();
 	fs_var_utmp();
@@ -1126,7 +1012,8 @@ void fs_overlayfs(void) {
 #endif
 
 
-#ifdef HAVE_CHROOT		
+#ifdef HAVE_CHROOT
+// return 1 if error
 void fs_check_chroot_dir(const char *rootdir) {
 	EUID_ASSERT();
 	assert(rootdir);
@@ -1173,7 +1060,7 @@ void fs_check_chroot_dir(const char *rootdir) {
 		exit(1);
 	}
 	free(name);
-	
+
 	// check /proc
 	if (asprintf(&name, "%s/proc", rootdir) == -1)
 		errExit("asprintf");
@@ -1186,7 +1073,7 @@ void fs_check_chroot_dir(const char *rootdir) {
 		exit(1);
 	}
 	free(name);
-	
+
 	// check /tmp
 	if (asprintf(&name, "%s/tmp", rootdir) == -1)
 		errExit("asprintf");
@@ -1230,7 +1117,6 @@ void fs_check_chroot_dir(const char *rootdir) {
 
 	// check x11 socket directory
 	if (getenv("FIREJAIL_X11")) {
-		mask_x11_abstract_socket = 1;
 		char *name;
 		if (asprintf(&name, "%s/tmp/.X11-unix", rootdir) == -1)
 			errExit("asprintf");
@@ -1249,7 +1135,7 @@ void fs_check_chroot_dir(const char *rootdir) {
 // chroot into an existing directory; mount exiting /dev and update /etc/resolv.conf
 void fs_chroot(const char *rootdir) {
 	assert(rootdir);
-	
+
 	if (checkcfg(CFG_CHROOT_DESKTOP)) {
 		// mount-bind a /dev in rootdir
 		char *newdev;
@@ -1260,10 +1146,9 @@ void fs_chroot(const char *rootdir) {
 		if (mount("/dev", newdev, NULL, MS_BIND|MS_REC, NULL) < 0)
 			errExit("mounting /dev");
 		free(newdev);
-		
+
 		// x11
 		if (getenv("FIREJAIL_X11")) {
-			mask_x11_abstract_socket = 1;
 			char *newx11;
 			if (asprintf(&newx11, "%s/tmp/.X11-unix", rootdir) == -1)
 				errExit("asprintf");
@@ -1273,10 +1158,10 @@ void fs_chroot(const char *rootdir) {
 				errExit("mounting /tmp/.X11-unix");
 			free(newx11);
 		}
-		
+
 		// some older distros don't have a /run directory
 		// create one by default
-		// no exit on error, let the user deal with any problems
+		// create /run/firejail directory in chroot
 		char *rundir;
 		if (asprintf(&rundir, "%s/run", rootdir) == -1)
 			errExit("asprintf");
@@ -1284,13 +1169,20 @@ void fs_chroot(const char *rootdir) {
 			fprintf(stderr, "Error: invalid run directory inside chroot\n");
 			exit(1);
 		}
-		if (!is_dir(rundir)) {
-			int rv = mkdir(rundir, 0755);
-			(void) rv;
-			rv = chown(rundir, 0, 0);
-			(void) rv;
-		}
-		
+		create_empty_dir_as_root(rundir, 0755);
+		free(rundir);
+		if (asprintf(&rundir, "%s/run/firejail", rootdir) == -1)
+			errExit("asprintf");
+		create_empty_dir_as_root(rundir, 0755);
+		free(rundir);
+
+		// create /run/firejail/mnt directory in chroot and mount the current one
+		if (asprintf(&rundir, "%s%s", rootdir, RUN_MNT_DIR) == -1)
+			errExit("asprintf");
+		create_empty_dir_as_root(rundir, 0755);
+		if (mount(RUN_MNT_DIR, rundir, NULL, MS_BIND|MS_REC, NULL) < 0)
+			errExit("mount bind");
+
 		// copy /etc/resolv.conf in chroot directory
 		// if resolv.conf in chroot is a symbolic link, this will fail
 		// no exit on error, let the user deal with the problem
@@ -1304,32 +1196,40 @@ void fs_chroot(const char *rootdir) {
 			exit(1);
 		}
 		if (copy_file("/etc/resolv.conf", fname, 0, 0, 0644) == -1) // root needed
-			fprintf(stderr, "Warning: /etc/resolv.conf not initialized\n");
+			fwarning("/etc/resolv.conf not initialized\n");
 	}
-	
+
 	// chroot into the new directory
+#ifdef HAVE_GCOV
+	__gcov_flush();
+#endif
 	if (arg_debug)
 		printf("Chrooting into %s\n", rootdir);
 	if (chroot(rootdir) < 0)
 		errExit("chroot");
-	// mount a new tmpfs in /run/firejail/mnt - the old one was lost in chroot
-	fs_build_remount_mnt_dir();
-		
+
+	// create all other /run/firejail files and directories
+	preproc_build_firejail_dir();
+
 	if (checkcfg(CFG_CHROOT_DESKTOP)) {
 		// update /var directory in order to support multiple sandboxes running on the same root directory
-		if (!arg_private_dev)
-			fs_dev_shm();
+//		if (!arg_private_dev)
+//			fs_dev_shm();
 		fs_var_lock();
 		fs_var_tmp();
-		fs_var_log();
+		if (!arg_writable_var_log)
+			fs_var_log();
+		else
+			fs_rdwr("/var/log");
+
 		fs_var_lib();
 		fs_var_cache();
 		fs_var_utmp();
 		fs_machineid();
-	
+
 		// don't leak user information
 		restrict_users();
-	
+
 		// when starting as root, firejail config is not disabled;
 		// this mode could be used to install and test new software by chaining
 		// firejail sandboxes (firejail --force)
@@ -1339,12 +1239,46 @@ void fs_chroot(const char *rootdir) {
 }
 #endif
 
+// this function is called from sandbox.c before blacklist/whitelist functions
 void fs_private_tmp(void) {
-	// mount tmpfs on top of /run/firejail/mnt
-	if (arg_debug)
-		printf("Mounting tmpfs on /tmp directory\n");
-	if (mount("tmpfs", "/tmp", "tmpfs", MS_NOSUID | MS_STRICTATIME | MS_REC,  "mode=1777,gid=0") < 0)
-		errExit("mounting tmpfs on /tmp directory");
-	fs_logger2("tmpfs", "/tmp");
-}
+	// check XAUTHORITY file, KDE keeps it under /tmp
+	char *xauth = getenv("XAUTHORITY");
+	if (xauth) {
+		char *rp = realpath(xauth, NULL);
+		if (rp && strncmp(rp, "/tmp/", 5) == 0) {
+			char *cmd;
+			if (asprintf(&cmd, "whitelist %s", rp) == -1)
+				errExit("asprintf");
+			profile_add(cmd); // profile_add does not duplicate the string
+		}
+		if (rp)
+			free(rp);
+	}
 
+	// whitelist x11 directory
+	profile_add("whitelist /tmp/.X11-unix");
+
+	// whitelist any pulse* file in /tmp directory
+	// some distros use PulseAudio sockets under /tmp instead of the socket in /urn/user
+	DIR *dir;
+	if (!(dir = opendir("/tmp"))) {
+		// sleep 2 seconds and try again
+		sleep(2);
+		if (!(dir = opendir("/tmp"))) {
+			return;
+		}
+	}
+
+	struct dirent *entry;
+	while ((entry = readdir(dir))) {
+		if (strncmp(entry->d_name, "pulse-", 6) == 0) {
+			char *cmd;
+			if (asprintf(&cmd, "whitelist /tmp/%s", entry->d_name) == -1)
+				errExit("asprintf");
+			profile_add(cmd); // profile_add does not duplicate the string
+		}
+	}
+	closedir(dir);
+
+
+}
