@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -37,12 +37,13 @@ static void mkdir_recursive(char *path) {
 	subdir = strtok(path, "/");
 	while(subdir) {
 		if (stat(subdir, &s) == -1) {
+			/* coverity[toctou] */
 			if (mkdir(subdir, 0700) == -1) {
-				fprintf(stderr, "Warning: cannot create %s directory\n", subdir);
+				fwarning("cannot create %s directory\n", subdir);
 				return;
 			}
 		} else if (!S_ISDIR(s.st_mode)) {
-			fprintf(stderr, "Warning: '%s' exists, but is no directory\n", subdir);
+			fwarning("'%s exists, but is not a directory\n", subdir);
 			return;
 		}
 		if (chdir(subdir)) {
@@ -56,12 +57,13 @@ static void mkdir_recursive(char *path) {
 
 void fs_mkdir(const char *name) {
 	EUID_ASSERT();
-	
+
 	// check directory name
 	invalid_filename(name);
 	char *expanded = expand_home(name, cfg.homedir);
-	if (strncmp(expanded, cfg.homedir, strlen(cfg.homedir)) != 0) {
-		fprintf(stderr, "Error: only directories in user home are supported by mkdir\n");
+	if (strncmp(expanded, cfg.homedir, strlen(cfg.homedir)) != 0 &&
+	    strncmp(expanded, "/tmp", 4) != 0) {
+		fprintf(stderr, "Error: only directories in user home or /tmp are supported by mkdir\n");
 		exit(1);
 	}
 
@@ -81,6 +83,9 @@ void fs_mkdir(const char *name) {
 
 		// create directory
 		mkdir_recursive(expanded);
+#ifdef HAVE_GCOV
+		__gcov_flush();
+#endif
 		_exit(0);
 	}
 	// wait for the child to finish
@@ -88,16 +93,17 @@ void fs_mkdir(const char *name) {
 
 doexit:
 	free(expanded);
-}	
+}
 
 void fs_mkfile(const char *name) {
 	EUID_ASSERT();
-	
+
 	// check file name
 	invalid_filename(name);
 	char *expanded = expand_home(name, cfg.homedir);
-	if (strncmp(expanded, cfg.homedir, strlen(cfg.homedir)) != 0) {
-		fprintf(stderr, "Error: only files in user home are supported by mkfile\n");
+	if (strncmp(expanded, cfg.homedir, strlen(cfg.homedir)) != 0 &&
+	    strncmp(expanded, "/tmp", 4) != 0) {
+		fprintf(stderr, "Error: only files in user home or /tmp are supported by mkfile\n");
 		exit(1);
 	}
 
