@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -34,8 +34,9 @@ static char *dentry[] = {
 	NULL
 };
 
+#define EMPTY_STRING ("")
 #define MAXBUF 4098
-static char *resolve_downloads(void) {
+static char *resolve_downloads(int nowhitelist_flag) {
 	char *fname;
 	struct stat s;
 
@@ -44,14 +45,20 @@ static char *resolve_downloads(void) {
 	while (dentry[i] != NULL) {
 		if (asprintf(&fname, "%s/%s", cfg.homedir, dentry[i]) == -1)
 			errExit("asprintf");
-		
+
 		if (stat(fname, &s) == 0) {
 			if (arg_debug || arg_debug_whitelists)
 				printf("Downloads directory resolved as \"%s\"\n", fname);
-			
+
 			char *rv;
-			if (asprintf(&rv, "whitelist ~/%s", dentry[i]) == -1)
-				errExit("asprintf");
+			if (nowhitelist_flag) {
+				if (asprintf(&rv, "nowhitelist ~/%s", dentry[i]) == -1)
+					errExit("asprintf");
+			}
+			else {
+				if (asprintf(&rv, "whitelist ~/%s", dentry[i]) == -1)
+					errExit("asprintf");
+			}
 			free(fname);
 			return rv;
 		}
@@ -66,14 +73,14 @@ static char *resolve_downloads(void) {
 	if (!fp) {
 		free(fname);
 		return NULL;
-	}		
+	}
 	free(fname);
-	
+
 	// extract downloads directory
 	char buf[MAXBUF];
 	while (fgets(buf, MAXBUF, fp)) {
 		char *ptr = buf;
-		
+
 		// skip blanks
 		while (*ptr == ' ' || *ptr == '\t')
 			ptr++;
@@ -91,45 +98,49 @@ static char *resolve_downloads(void) {
 				if (strlen(ptr1) != 0) {
 					if (arg_debug || arg_debug_whitelists)
 						printf("Downloads directory resolved as \"%s\"\n", ptr1);
-				
+
 					if (asprintf(&fname, "%s/%s", cfg.homedir, ptr1) == -1)
 						errExit("asprintf");
-					
+
 					if (stat(fname, &s) == -1) {
-						fprintf(stderr, "***\n");
-						fprintf(stderr, "*** Error: directory %s not found.\n", fname);
-						fprintf(stderr, "*** \tThis directory is configured in ~/.config/user-dirs.dirs.\n");
-						fprintf(stderr, "*** \tPlease create a Downloads directory.\n");
-						fprintf(stderr, "***\n");
 						free(fname);
-						return NULL;
+						goto errout;
 					}
-				
+
 					char *rv;
-					if (asprintf(&rv, "whitelist ~/%s", ptr + 24) == -1)
-						errExit("asprintf");
+					if (nowhitelist_flag) {
+						if (asprintf(&rv, "nowhitelist ~/%s", ptr + 24) == -1)
+							errExit("asprintf");
+					}
+					else {
+						if (asprintf(&rv, "whitelist ~/%s", ptr + 24) == -1)
+							errExit("asprintf");
+					}
 					return rv;
 				}
-				else {
-					fprintf(stderr, "***\n");
-					fprintf(stderr, "*** Error: invalid XDG_DOWNLOAD_DIR entry in ~/.config/user-dirs.dirs.\n");
-					fprintf(stderr, "*** \tPlease specify a valid Downloads directory, example:\n");
-					fprintf(stderr, "***\n");
-					fprintf(stderr, "***\t\tXDG_DOWNLOAD_DIR=\"$HOME/Downloads\"\n");
-					fprintf(stderr, "***\n");
-					return NULL;
-				}
+				else
+					goto errout;
 			}
 		}
 	}
+
 	fclose(fp);
-	
+	return NULL;
+
+errout:
+	if (!arg_private) {
+		fprintf(stderr, "***\n");
+		fprintf(stderr, "*** Error: Downloads directory was not found in user home.\n");
+		fprintf(stderr, "*** \tAny files saved by the program, will be lost when the sandbox is closed.\n");
+		fprintf(stderr, "***\n");
+	}
+
 	return NULL;
 }
 
 static int mkpath(const char* path, mode_t mode) {
 	assert(path && *path);
-	
+
 	mode |= 0111;
 
 	// create directories with uid/gid as root or as current user if inside home directory
@@ -157,18 +168,16 @@ static int mkpath(const char* path, mode_t mode) {
 			}
 		}
 		else {
-			if (chmod(file_path, mode) == -1)
-				errExit("chmod");
-			if (chown(file_path, uid, gid) == -1)
-				errExit("chown");
+			if (set_perms(file_path, uid, gid, mode))
+				errExit("set_perms");
 			done = 1;
-		}			
+		}
 
 		*p='/';
 	}
 	if (done)
 		fs_logger2("mkpath", path);
-	
+
 	free(file_path);
 	return 0;
 }
@@ -181,85 +190,69 @@ static void whitelist_path(ProfileEntry *entry) {
 	char *wfile = NULL;
 
 	if (entry->home_dir) {
-		if (strncmp(path, cfg.homedir, strlen(cfg.homedir)) == 0) {		
+		if (strncmp(path, cfg.homedir, strlen(cfg.homedir)) == 0) {
 			fname = path + strlen(cfg.homedir);
-			if (*fname == '\0') {
-				fprintf(stderr, "Error: file %s is not in user home directory, exiting...\n", path);
-				exit(1);
-			}
+			if (*fname == '\0')
+				goto errexit;
 		}
 		else
 			fname = path;
-	
+
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_HOME_USER_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->tmp_dir) {
 		fname = path + 4; // strlen("/tmp")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /tmp directory, exiting...\n", path);
-			exit(1);
-		}
-	
+		if (*fname == '\0')
+			goto errexit;
+
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_TMP_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->media_dir) {
 		fname = path + 6; // strlen("/media")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /media directory, exiting...\n", path);
-			exit(1);
-		}
-	
+		if (*fname == '\0')
+			goto errexit;
+
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_MEDIA_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->mnt_dir) {
 		fname = path + 4; // strlen("/mnt")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /mnt directory, exiting...\n", path);
-			exit(1);
-		}
+		if (*fname == '\0')
+			goto errexit;
 
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_MNT_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->var_dir) {
 		fname = path + 4; // strlen("/var")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /var directory, exiting...\n", path);
-			exit(1);
-		}
-	
+		if (*fname == '\0')
+			goto errexit;
+
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_VAR_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->dev_dir) {
 		fname = path + 4; // strlen("/dev")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /dev directory, exiting...\n", path);
-			exit(1);
-		}
-	
+		if (*fname == '\0')
+			goto errexit;
+
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_DEV_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->opt_dir) {
 		fname = path + 4; // strlen("/opt")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /opt directory, exiting...\n", path);
-			exit(1);
-		}
-	
+		if (*fname == '\0')
+			goto errexit;
+
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_OPT_DIR, fname) == -1)
 			errExit("asprintf");
 	}
 	else if (entry->srv_dir) {
 		fname = path + 4; // strlen("/srv")
-		if (*fname == '\0') {
-			fprintf(stderr, "Error: file %s is not in /srv directory, exiting...\n", path);
-			exit(1);
-		}
+		if (*fname == '\0')
+			goto errexit;
 
 		if (asprintf(&wfile, "%s/%s", RUN_WHITELIST_SRV_DIR, fname) == -1)
 			errExit("asprintf");
@@ -273,18 +266,18 @@ static void whitelist_path(ProfileEntry *entry) {
 	else {
 		return;
 	}
-	
+
 	// create the path if necessary
 	mkpath(path, s.st_mode);
 	fs_logger2("whitelist", path);
 
 	// process directory
-	if (S_ISDIR(s.st_mode)) {	
+	if (S_ISDIR(s.st_mode)) {
 		// create directory
 		int rv = mkdir(path, 0755);
 		(void) rv;
 	}
-	
+
 	// process regular file
 	else {
 		if (access(path, R_OK)) {
@@ -301,12 +294,17 @@ static void whitelist_path(ProfileEntry *entry) {
 		else
 			return; // the file is already present
 	}
-	
+
 	// mount
 	if (mount(wfile, path, NULL, MS_BIND|MS_REC, NULL) < 0)
 		errExit("mount bind");
 
 	free(wfile);
+	return;
+
+errexit:
+	fprintf(stderr, "Error: file %s is not in the whitelisted directory\n", path);
+	exit(1);
 }
 
 
@@ -326,40 +324,58 @@ void fs_whitelist(void) {
 	int var_dir = 0;		// /var directory flag
 	int dev_dir = 0;		// /dev directory flag
 	int opt_dir = 0;		// /opt directory flag
-        int srv_dir = 0;                // /srv directory flag
+	int srv_dir = 0;                // /srv directory flag
+
+	size_t nowhitelist_c = 0;
+	size_t nowhitelist_m = 32;
+	char **nowhitelist = calloc(nowhitelist_m, sizeof(*nowhitelist));
+	if (nowhitelist == NULL)
+		errExit("failed allocating memory for nowhitelist entries");
+
 	// verify whitelist files, extract symbolic links, etc.
 	while (entry) {
-		// handle only whitelist commands
-		if (strncmp(entry->data, "whitelist ", 10)) {
+		int nowhitelist_flag = 0;
+
+		// handle only whitelist and nowhitelist commands
+		if (strncmp(entry->data, "whitelist ", 10) == 0)
+			nowhitelist_flag = 0;
+		else if (strncmp(entry->data, "nowhitelist ", 12) == 0)
+			nowhitelist_flag = 1;
+		else {
 			entry = entry->next;
 			continue;
 		}
+		char *dataptr = (nowhitelist_flag)? entry->data + 12: entry->data + 10;
 
 		// resolve ${DOWNLOADS}
-		if (strcmp(entry->data + 10, "${DOWNLOADS}") == 0) {
-			char *tmp = resolve_downloads();
-			if (tmp)
+		if (strcmp(dataptr, "${DOWNLOADS}") == 0) {
+			char *tmp = resolve_downloads(nowhitelist_flag);
+			if (tmp) {
 				entry->data = tmp;
+				dataptr = (nowhitelist_flag)? entry->data + 12: entry->data + 10;
+			}
 			else {
-				*entry->data = '\0';
-				fprintf(stderr, "***\n");
-				fprintf(stderr, "*** Warning: cannot whitelist Downloads directory\n");
-				fprintf(stderr, "*** \tAny file saved will be lost when the sandbox is closed.\n");
-				fprintf(stderr, "*** \tPlease create a proper Downloads directory for your application.\n");
-				fprintf(stderr, "***\n");
+				if (!nowhitelist_flag && !arg_quiet && !arg_private) {
+					fprintf(stderr, "***\n");
+					fprintf(stderr, "*** Warning: cannot whitelist Downloads directory\n");
+					fprintf(stderr, "*** \tAny file saved will be lost when the sandbox is closed.\n");
+					fprintf(stderr, "*** \tPlease create a proper Downloads directory for your application.\n");
+					fprintf(stderr, "***\n");
+				}
+				entry->data = EMPTY_STRING;
 				continue;
 			}
 		}
 
 		// replace ~/ or ${HOME} into /home/username
-		new_name = expand_home(entry->data + 10, cfg.homedir);
+		new_name = expand_home(dataptr, cfg.homedir);
 		assert(new_name);
-		if (arg_debug)
-			fprintf(stderr, "Debug %d: new_name #%s#\n", __LINE__, new_name);
+		if (arg_debug || arg_debug_whitelists)
+			fprintf(stderr, "Debug %d: new_name #%s#, %s\n", __LINE__, new_name, (nowhitelist_flag)? "nowhitelist": "whitelist");
 
 		// valid path referenced to filesystem root
 		if (*new_name != '/') {
-			if (arg_debug)
+			if (arg_debug || arg_debug_whitelists)
 				fprintf(stderr, "Debug %d: \n", __LINE__);
 			goto errexit;
 		}
@@ -371,45 +387,66 @@ void fs_whitelist(void) {
 		if (!fname) {
 			// file not found, blank the entry in the list and continue
 			if (arg_debug || arg_debug_whitelists) {
-				printf("Removed whitelist path: %s\n", entry->data);
+				printf("Removed whitelist/nowhitelist path: %s\n", entry->data);
 				printf("\texpanded: %s\n", new_name);
 				printf("\treal path: (null)\n");
 				printf("\t");fflush(0);
 				perror("realpath");
 			}
-			*entry->data = '\0';
 
 			// if 1 the file was not found; mount an empty directory
-			if (strncmp(new_name, cfg.homedir, strlen(cfg.homedir)) == 0) {
-				if(!arg_private)
-					home_dir = 1;
+			if (!nowhitelist_flag) {
+				if (strncmp(new_name, cfg.homedir, strlen(cfg.homedir)) == 0) {
+					if(!arg_private)
+						home_dir = 1;
+				}
+				else if (strncmp(new_name, "/tmp/", 5) == 0)
+					tmp_dir = 1;
+				else if (strncmp(new_name, "/media/", 7) == 0)
+					media_dir = 1;
+				else if (strncmp(new_name, "/mnt/", 5) == 0)
+					mnt_dir = 1;
+				else if (strncmp(new_name, "/var/", 5) == 0)
+					var_dir = 1;
+				else if (strncmp(new_name, "/dev/", 5) == 0)
+					dev_dir = 1;
+				else if (strncmp(new_name, "/opt/", 5) == 0)
+					opt_dir = 1;
+				else if (strncmp(new_name, "/srv/", 5) == 0)
+					opt_dir = 1;
 			}
-			else if (strncmp(new_name, "/tmp/", 5) == 0)
-				tmp_dir = 1;
-			else if (strncmp(new_name, "/media/", 7) == 0)
-				media_dir = 1;
-			else if (strncmp(new_name, "/mnt/", 5) == 0)
-				mnt_dir = 1;
-			else if (strncmp(new_name, "/var/", 5) == 0)
-				var_dir = 1;
-			else if (strncmp(new_name, "/dev/", 5) == 0)
-				dev_dir = 1;
-			else if (strncmp(new_name, "/opt/", 5) == 0)
-				opt_dir = 1;
-			else if (strncmp(new_name, "/srv/", 5) == 0)
-				opt_dir = 1;
-			
+
+			entry->data = EMPTY_STRING;
 			continue;
 		}
-		
+		else if (arg_debug_whitelists)
+			printf("real path %s\n", fname);
+
+		if (nowhitelist_flag) {
+			// store the path in nowhitelist array
+			if (arg_debug || arg_debug_whitelists)
+				printf("Storing nowhitelist %s\n", fname);
+
+			if (nowhitelist_c >= nowhitelist_m) {
+				nowhitelist_m *= 2;
+				nowhitelist = realloc(nowhitelist, sizeof(*nowhitelist) * nowhitelist_m);
+				if (nowhitelist == NULL)
+					errExit("failed increasing memory for nowhitelist entries");
+			}
+			nowhitelist[nowhitelist_c++] = fname;
+			entry->data = EMPTY_STRING;
+			continue;
+		}
+
+
 		// check for supported directories
 		if (strncmp(new_name, cfg.homedir, strlen(cfg.homedir)) == 0) {
 			// whitelisting home directory is disabled if --private option is present
 			if (arg_private) {
 				if (arg_debug || arg_debug_whitelists)
-					printf("Removed whitelist path %s, --private option is present\n", entry->data);
+					printf("\"%s\" disabled by --private\n", entry->data);
 
-				*entry->data = '\0';
+				entry->data = EMPTY_STRING;
 				continue;
 			}
 
@@ -421,10 +458,12 @@ void fs_whitelist(void) {
 
 			// both path and absolute path are under /home
 			if (strncmp(fname, cfg.homedir, strlen(cfg.homedir)) != 0) {
-				// check if the file is owned by the user
-				struct stat s;
-				if (stat(fname, &s) == 0 && s.st_uid != getuid())
-					goto errexit;
+				if (checkcfg(CFG_FOLLOW_SYMLINK_AS_USER)) {
+					// check if the file is owned by the user
+					struct stat s;
+					if (stat(fname, &s) == 0 && s.st_uid != getuid())
+						goto errexit;
+				}
 			}
 		}
 		else if (strncmp(new_name, "/tmp/", 5) == 0) {
@@ -432,8 +471,6 @@ void fs_whitelist(void) {
 			tmp_dir = 1;
 			// both path and absolute path are under /tmp
 			if (strncmp(fname, "/tmp/", 5) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
 				goto errexit;
 			}
 		}
@@ -442,8 +479,6 @@ void fs_whitelist(void) {
 			media_dir = 1;
 			// both path and absolute path are under /media
 			if (strncmp(fname, "/media/", 7) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
 				goto errexit;
 			}
 		}
@@ -452,8 +487,6 @@ void fs_whitelist(void) {
 			mnt_dir = 1;
 			// both path and absolute path are under /mnt
 			if (strncmp(fname, "/mnt/", 5) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
 				goto errexit;
 			}
 		}
@@ -467,19 +500,21 @@ void fs_whitelist(void) {
 			else if (strcmp(new_name, "/var/lock")== 0)
 				;
 			else if (strncmp(fname, "/var/", 5) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
 				goto errexit;
 			}
 		}
 		else if (strncmp(new_name, "/dev/", 5) == 0) {
 			entry->dev_dir = 1;
 			dev_dir = 1;
-			// both path and absolute path are under /dev
-			if (strncmp(fname, "/dev/", 5) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
-				goto errexit;
+
+			// special handling for /dev/shm
+			// on some platforms (Debian wheezy, Ubuntu 14.04), it is a symlink to /run/shm
+			if (strcmp(new_name, "/dev/shm") == 0 && strcmp(fname, "/run/shm") == 0);
+			else {
+				// both path and absolute path are under /dev
+				if (strncmp(fname, "/dev/", 5) != 0) {
+					goto errexit;
+				}
 			}
 		}
 		else if (strncmp(new_name, "/opt/", 5) == 0) {
@@ -487,8 +522,6 @@ void fs_whitelist(void) {
 			opt_dir = 1;
 			// both path and absolute path are under /dev
 			if (strncmp(fname, "/opt/", 5) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
 				goto errexit;
 			}
 		}
@@ -497,17 +530,34 @@ void fs_whitelist(void) {
 			srv_dir = 1;
 			// both path and absolute path are under /srv
 			if (strncmp(fname, "/srv/", 5) != 0) {
-				if (arg_debug)
-					fprintf(stderr, "Debug %d: fname #%s#\n", __LINE__, fname);
 				goto errexit;
 			}
 		}
 		else {
-			if (arg_debug)
-				fprintf(stderr, "Debug %d: \n", __LINE__);
 			goto errexit;
 		}
 
+		// check if the path is in nowhitelist array
+		if (nowhitelist_flag == 0) {
+			size_t i;
+			int found = 0;
+			for (i = 0; i < nowhitelist_c; i++) {
+				if (nowhitelist[i] == NULL)
+					break;
+				if (strcmp(nowhitelist[i], fname) == 0) {
+					found = 1;
+					break;
+				}
+			}
+			if (found) {
+				if (arg_debug || arg_debug_whitelists)
+					printf("Skip nowhitelisted path %s\n", fname);
+				entry->data = EMPTY_STRING;
+				free(fname);
+				continue;
+			}
+		}
+
 		// mark symbolic links
 		if (is_link(new_name))
 			entry->link = new_name;
@@ -528,43 +578,29 @@ void fs_whitelist(void) {
 		free(fname);
 		entry = entry->next;
 	}
-		
-	// create mount points
-	fs_build_mnt_dir();
-	
+
+	// release nowhitelist memory
+	assert(nowhitelist);
+	free(nowhitelist);
 
 	// /home/user
 	if (home_dir) {
 		// keep a copy of real home dir in RUN_WHITELIST_HOME_USER_DIR
-		int rv = mkdir(RUN_WHITELIST_HOME_USER_DIR, 0755);
-		if (rv == -1)
-			errExit("mkdir");
-		if (chown(RUN_WHITELIST_HOME_USER_DIR, getuid(), getgid()) < 0)
-			errExit("chown");
-		if (chmod(RUN_WHITELIST_HOME_USER_DIR, 0755) < 0)
-			errExit("chmod");
-	
+		mkdir_attr(RUN_WHITELIST_HOME_USER_DIR, 0755, getuid(), getgid());
 		if (mount(cfg.homedir, RUN_WHITELIST_HOME_USER_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 			errExit("mount bind");
-	
+
 		// mount a tmpfs and initialize /home/user
 		 fs_private();
 	}
-	
+
 	// /tmp mountpoint
 	if (tmp_dir) {
-		// keep a copy of real /tmp directory in  
-		int rv = mkdir(RUN_WHITELIST_TMP_DIR, 1777);
-		if (rv == -1)
-			errExit("mkdir");
-		if (chown(RUN_WHITELIST_TMP_DIR, 0, 0) < 0)
-			errExit("chown");
-		if (chmod(RUN_WHITELIST_TMP_DIR, 1777) < 0)
-			errExit("chmod");
-	
+		// keep a copy of real /tmp directory in
+		mkdir_attr(RUN_WHITELIST_TMP_DIR, 1777, 0, 0);
 		if (mount("/tmp", RUN_WHITELIST_TMP_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 			errExit("mount bind");
-	
+
 		// mount tmpfs on /tmp
 		if (arg_debug || arg_debug_whitelists)
 			printf("Mounting tmpfs on /tmp directory\n");
@@ -572,24 +608,17 @@ void fs_whitelist(void) {
 			errExit("mounting tmpfs on /tmp");
 		fs_logger("tmpfs /tmp");
 	}
-	
+
 	// /media mountpoint
 	if (media_dir) {
 		// some distros don't have a /media directory
 		struct stat s;
 		if (stat("/media", &s) == 0) {
 			// keep a copy of real /media directory in RUN_WHITELIST_MEDIA_DIR
-			int rv = mkdir(RUN_WHITELIST_MEDIA_DIR, 0755);
-			if (rv == -1)
-				errExit("mkdir");
-			if (chown(RUN_WHITELIST_MEDIA_DIR, 0, 0) < 0)
-				errExit("chown");
-			if (chmod(RUN_WHITELIST_MEDIA_DIR, 0755) < 0)
-				errExit("chmod");
-	
+			mkdir_attr(RUN_WHITELIST_MEDIA_DIR, 0755, 0, 0);
 			if (mount("/media", RUN_WHITELIST_MEDIA_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 				errExit("mount bind");
-	
+
 			// mount tmpfs on /media
 			if (arg_debug || arg_debug_whitelists)
 				printf("Mounting tmpfs on /media directory\n");
@@ -607,14 +636,7 @@ void fs_whitelist(void) {
 		struct stat s;
 		if (stat("/mnt", &s) == 0) {
 			// keep a copy of real /mnt directory in RUN_WHITELIST_MNT_DIR
-			int rv = mkdir(RUN_WHITELIST_MNT_DIR, 0755);
-			if (rv == -1)
-				errExit("mkdir");
-			if (chown(RUN_WHITELIST_MNT_DIR, 0, 0) < 0)
-				errExit("chown");
-			if (chmod(RUN_WHITELIST_MNT_DIR, 0755) < 0)
-				errExit("chmod");
-
+			mkdir_attr(RUN_WHITELIST_MNT_DIR, 0755, 0, 0);
 			if (mount("/mnt", RUN_WHITELIST_MNT_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 				errExit("mount bind");
 
@@ -633,17 +655,10 @@ void fs_whitelist(void) {
 	// /var mountpoint
 	if (var_dir) {
 		// keep a copy of real /var directory in RUN_WHITELIST_VAR_DIR
-		int rv = mkdir(RUN_WHITELIST_VAR_DIR, 0755);
-		if (rv == -1)
-			errExit("mkdir");
-		if (chown(RUN_WHITELIST_VAR_DIR, 0, 0) < 0)
-			errExit("chown");
-		if (chmod(RUN_WHITELIST_VAR_DIR, 0755) < 0)
-			errExit("chmod");
-	
+		mkdir_attr(RUN_WHITELIST_VAR_DIR, 0755, 0, 0);
 		if (mount("/var", RUN_WHITELIST_VAR_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 			errExit("mount bind");
-	
+
 		// mount tmpfs on /var
 		if (arg_debug || arg_debug_whitelists)
 			printf("Mounting tmpfs on /var directory\n");
@@ -655,17 +670,10 @@ void fs_whitelist(void) {
 	// /dev mountpoint
 	if (dev_dir) {
 		// keep a copy of real /dev directory in RUN_WHITELIST_DEV_DIR
-		int rv = mkdir(RUN_WHITELIST_DEV_DIR, 0755);
-		if (rv == -1)
-			errExit("mkdir");
-		if (chown(RUN_WHITELIST_DEV_DIR, 0, 0) < 0)
-			errExit("chown");
-		if (chmod(RUN_WHITELIST_DEV_DIR, 0755) < 0)
-			errExit("chmod");
-	
+		mkdir_attr(RUN_WHITELIST_DEV_DIR, 0755, 0, 0);
 		if (mount("/dev", RUN_WHITELIST_DEV_DIR, NULL, MS_BIND|MS_REC,  "mode=755,gid=0") < 0)
 			errExit("mount bind");
-	
+
 		// mount tmpfs on /dev
 		if (arg_debug || arg_debug_whitelists)
 			printf("Mounting tmpfs on /dev directory\n");
@@ -677,17 +685,10 @@ void fs_whitelist(void) {
 	// /opt mountpoint
 	if (opt_dir) {
 		// keep a copy of real /opt directory in RUN_WHITELIST_OPT_DIR
-		int rv = mkdir(RUN_WHITELIST_OPT_DIR, 0755);
-		if (rv == -1)
-			errExit("mkdir");
-		if (chown(RUN_WHITELIST_OPT_DIR, 0, 0) < 0)
-			errExit("chown");
-		if (chmod(RUN_WHITELIST_OPT_DIR, 0755) < 0)
-			errExit("chmod");
-	
+		mkdir_attr(RUN_WHITELIST_OPT_DIR, 0755, 0, 0);
 		if (mount("/opt", RUN_WHITELIST_OPT_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 			errExit("mount bind");
-	
+
 		// mount tmpfs on /opt
 		if (arg_debug || arg_debug_whitelists)
 			printf("Mounting tmpfs on /opt directory\n");
@@ -702,14 +703,7 @@ void fs_whitelist(void) {
 		struct stat s;
 		if (stat("/srv", &s) == 0) {
 			// keep a copy of real /srv directory in RUN_WHITELIST_SRV_DIR
-			int rv = mkdir(RUN_WHITELIST_SRV_DIR, 0755);
-			if (rv == -1)
-				errExit("mkdir");
-			if (chown(RUN_WHITELIST_SRV_DIR, 0, 0) < 0)
-				errExit("chown");
-			if (chmod(RUN_WHITELIST_SRV_DIR, 0755) < 0)
-				errExit("chmod");
-
+			mkdir_attr(RUN_WHITELIST_SRV_DIR, 0755, 0, 0);
 			if (mount("/srv", RUN_WHITELIST_SRV_DIR, NULL, MS_BIND|MS_REC, NULL) < 0)
 				errExit("mount bind");
 
@@ -725,7 +719,6 @@ void fs_whitelist(void) {
 	}
 
 
-	
 	// go through profile rules again, and interpret whitelist commands
 	entry = cfg.profile;
 	while (entry) {
@@ -737,7 +730,7 @@ void fs_whitelist(void) {
 
 //printf("here %d#%s#\n", __LINE__, entry->data);
 		// whitelist the real file
-		if (strcmp(entry->data, "whitelist /run") == 0 && 
+		if (strcmp(entry->data, "whitelist /run") == 0 &&
 		    (strcmp(entry->link, "/var/run") == 0 || strcmp(entry->link, "/var/lock") == 0)) {
 			int rv = symlink(entry->data + 10, entry->link);
 			if (rv)
@@ -747,7 +740,7 @@ void fs_whitelist(void) {
 		}
 		else {
 			whitelist_path(entry);
-	
+
 			// create the link if any
 			if (entry->link) {
 				// if the link is already there, do not bother
@@ -755,7 +748,7 @@ void fs_whitelist(void) {
 				if (stat(entry->link, &s) != 0) {
 					// create the path if necessary
 					mkpath(entry->link, s.st_mode);
-	
+
 					int rv = symlink(entry->data + 10, entry->link);
 					if (rv)
 						fprintf(stderr, "Warning cannot create symbolic link %s\n", entry->link);
@@ -774,7 +767,7 @@ void fs_whitelist(void) {
 			errExit("mount tmpfs");
 		fs_logger2("tmpfs", RUN_WHITELIST_HOME_USER_DIR);
 	}
-	
+
 	// mask the real /tmp directory, currently mounted on RUN_WHITELIST_TMP_DIR
 	if (tmp_dir) {
 		if (mount("tmpfs", RUN_WHITELIST_TMP_DIR, "tmpfs", MS_NOSUID | MS_STRICTATIME | MS_REC,  "mode=755,gid=0") < 0)
@@ -826,7 +819,7 @@ void fs_whitelist(void) {
 
 	if (new_name)
 		free(new_name);
-		
+
 	return;
 
 errexit:
