@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -28,6 +28,7 @@
 static int apply_caps = 0;
 static uint64_t caps = 0;
 static int apply_seccomp = 0;
+static unsigned display = 0;
 #define BUFLEN 4096
 
 static void signal_handler(int sig){
@@ -36,6 +37,37 @@ static void signal_handler(int sig){
 	exit(sig);
 }
 
+
+
+static void extract_x11_display(pid_t pid) {
+	char *fname;
+	if (asprintf(&fname, "%s/%d", RUN_FIREJAIL_X11_DIR, pid) == -1)
+		errExit("asprintf");
+
+	FILE *fp = fopen(fname, "r");
+	free(fname);
+	if (!fp)
+		return;
+
+	if (1 != fscanf(fp, "%d", &display)) {
+		fprintf(stderr, "Error: cannot read X11 display file\n");
+		fclose(fp);
+		return;
+	}
+	fclose(fp);
+
+	// check display range
+	if (display < X11_DISPLAY_START || display > X11_DISPLAY_END) {
+		fprintf(stderr, "Error: invalid X11 display range\n");
+		return;
+	}
+
+	// store the display number for join process in /run/firejail/x11
+	EUID_ROOT();
+	set_x11_file(getpid(), display);
+	EUID_USER();
+}
+
 static void extract_command(int argc, char **argv, int index) {
 	EUID_ASSERT();
 	if (index >= argc)
@@ -48,7 +80,7 @@ static void extract_command(int argc, char **argv, int index) {
 		if (index >= argc)
 			return;
 	}
-		
+
 	// first argv needs to be a valid command
 	if (arg_doubledash == 0 && *argv[index] == '-') {
 		fprintf(stderr, "Error: invalid option %s after --join\n", argv[index]);
@@ -66,7 +98,7 @@ static void extract_nogroups(pid_t pid) {
 	char *fname;
 	if (asprintf(&fname, "/proc/%d/root%s", pid, RUN_GROUPS_CFG) == -1)
 		errExit("asprintf");
-		
+
 	struct stat s;
 	if (stat(fname, &s) == -1)
 		return;
@@ -79,11 +111,11 @@ static void extract_cpu(pid_t pid) {
 	char *fname;
 	if (asprintf(&fname, "/proc/%d/root%s", pid, RUN_CPU_CFG) == -1)
 		errExit("asprintf");
-		
+
 	struct stat s;
 	if (stat(fname, &s) == -1)
 		return;
-	
+
 	// there is a CPU_CFG file, load it!
 	load_cpu(fname);
 	free(fname);
@@ -93,11 +125,11 @@ static void extract_cgroup(pid_t pid) {
 	char *fname;
 	if (asprintf(&fname, "/proc/%d/root%s", pid, RUN_CGROUP_CFG) == -1)
 		errExit("asprintf");
-		
+
 	struct stat s;
 	if (stat(fname, &s) == -1)
 		return;
-	
+
 	// there is a cgroup file CGROUP_CFG, load it!
 	load_cgroup(fname);
 	free(fname);
@@ -127,7 +159,7 @@ static void extract_caps_seccomp(pid_t pid) {
 				apply_seccomp = 1;
 			break;
 		}
-		else if (strncmp(buf, "CapBnd:", 7) == 0) {		
+		else if (strncmp(buf, "CapBnd:", 7) == 0) {
 			char *ptr = buf + 7;
 			unsigned long long val;
 			sscanf(ptr, "%llx", &val);
@@ -149,7 +181,7 @@ static void extract_user_namespace(pid_t pid) {
 	    stat("/proc/self/gid_map", &s3) == 0);
 	else
 		return;
-			
+
 	// read uid map
 	char *uidmap;
 	if (asprintf(&uidmap, "/proc/%u/uid_map", pid) == -1)
@@ -173,25 +205,11 @@ static void extract_user_namespace(pid_t pid) {
 	free(uidmap);
 }
 
-void join_name(const char *name, int argc, char **argv, int index) {
-	EUID_ASSERT();
-	if (!name || strlen(name) == 0) {
-		fprintf(stderr, "Error: invalid sandbox name\n");
-		exit(1);
-	}
-
-	pid_t pid;
-	if (name2pid(name, &pid)) {
-		fprintf(stderr, "Error: cannot find sandbox %s\n", name);
-		exit(1);
-	}
-	join(pid, argc, argv, index);
-}
-
 void join(pid_t pid, int argc, char **argv, int index) {
 	EUID_ASSERT();
 	char *homedir = cfg.homedir;
-	
+	pid_t parent = pid;
+
 	extract_command(argc, argv, index);
 	signal (SIGTERM, signal_handler);
 
@@ -204,7 +222,8 @@ void join(pid_t pid, int argc, char **argv, int index) {
 			pid_t child;
 			if (find_child(pid, &child) == 0) {
 				pid = child;
-				printf("Switching to pid %u, the first child process inside the sandbox\n", (unsigned) pid);
+				if (!arg_quiet)
+					printf("Switching to pid %u, the first child process inside the sandbox\n", (unsigned) pid);
 			}
 		}
 		free(comm);
@@ -220,6 +239,8 @@ void join(pid_t pid, int argc, char **argv, int index) {
 		}
 	}
 
+	extract_x11_display(parent);
+
 	EUID_ROOT();
 	// in user mode set caps seccomp, cpu, cgroup, etc
 	if (getuid() != 0) {
@@ -229,11 +250,11 @@ void join(pid_t pid, int argc, char **argv, int index) {
 		extract_nogroups(pid);
 		extract_user_namespace(pid);
 	}
-	
+
 	// set cgroup
 	if (cfg.cgroup)	// not available for uid 0
 		set_cgroup(cfg.cgroup);
-		
+
 	// join namespaces
 	if (arg_join_network) {
 		if (join_namespace(pid, "net"))
@@ -244,15 +265,11 @@ void join(pid_t pid, int argc, char **argv, int index) {
 			exit(1);
 	}
 	else {
-		if (join_namespace(pid, "ipc"))
-			exit(1);
-		if (join_namespace(pid, "net"))
-			exit(1);
-		if (join_namespace(pid, "pid"))
-			exit(1);
-		if (join_namespace(pid, "uts"))
-			exit(1);
-		if (join_namespace(pid, "mnt"))
+		if (join_namespace(pid, "ipc") ||
+		    join_namespace(pid, "net") ||
+		    join_namespace(pid, "pid") ||
+		    join_namespace(pid, "uts") ||
+		    join_namespace(pid, "mnt"))
 			exit(1);
 	}
 
@@ -260,18 +277,21 @@ void join(pid_t pid, int argc, char **argv, int index) {
 	if (child < 0)
 		errExit("fork");
 	if (child == 0) {
+		// drop discretionary access control capabilities for root sandboxes
+		caps_drop_dac_override();
+		
 		// chroot into /proc/PID/root directory
 		char *rootdir;
 		if (asprintf(&rootdir, "/proc/%d/root", pid) == -1)
 			errExit("asprintf");
-			
+
 		int rv;
 		if (!arg_join_network) {
 			rv = chroot(rootdir); // this will fail for processes in sandboxes not started with --chroot option
 			if (rv == 0)
 				printf("changing root to %s\n", rootdir);
 		}
-		
+
 		prctl(PR_SET_PDEATHSIG, SIGKILL, 0, 0, 0); // kill the child in case the parent died
 		if (chdir("/") < 0)
 			errExit("chdir");
@@ -283,33 +303,27 @@ void join(pid_t pid, int argc, char **argv, int index) {
 					errExit("chdir");
 			}
 		}
-		
+
 		// set cpu affinity
 		if (cfg.cpus)	// not available for uid 0
 			set_cpu_affinity();
-					
+
 		// set caps filter
 		if (apply_caps == 1)	// not available for uid 0
 			caps_set(caps);
 #ifdef HAVE_SECCOMP
-		// set protocol filter
+		// read cfg.protocol from file
 		if (getuid() != 0)
 			protocol_filter_load(RUN_PROTOCOL_CFG);
 		if (cfg.protocol) {	// not available for uid 0
-			protocol_filter();
+			seccomp_load(RUN_SECCOMP_PROTOCOL);	// install filter
 		}
-				
+
 		// set seccomp filter
 		if (apply_seccomp == 1)	// not available for uid 0
-			seccomp_set();
+			seccomp_load(RUN_SECCOMP_CFG);
 #endif
 
-		// fix qt 4.8
-		if (setenv("QT_X11_NO_MITSHM", "1", 1) < 0)
-			errExit("setenv");
-		if (setenv("container", "firejail", 1) < 0) // LXC sets container=lxc,
-			errExit("setenv");
-
 		// mount user namespace or drop privileges
 		if (arg_noroot) {	// not available for uid 0
 			if (arg_debug)
@@ -322,47 +336,33 @@ void join(pid_t pid, int argc, char **argv, int index) {
 			if (apply_caps == 1)	// not available for uid 0
 				caps_set(caps);
 		}
-		else 
+		else
 			drop_privs(arg_nogroups);	// nogroups not available for uid 0
 
 
-		// set prompt color to green
-		char *prompt = getenv("FIREJAIL_PROMPT");
-		if (prompt && strcmp(prompt, "yes") == 0) {
-			//export PS1='\[\e[1;32m\][\u@\h \W]\$\[\e[0m\] '
-			if (setenv("PROMPT_COMMAND", "export PS1=\"\\[\\e[1;32m\\][\\u@\\h \\W]\\$\\[\\e[0m\\] \"", 1) < 0)
-				errExit("setenv");
-		}
-		
 		// set nice
 		if (arg_nice) {
 			errno = 0;
 			int rv = nice(cfg.nice);
 			(void) rv;
 			if (errno) {
-				fprintf(stderr, "Warning: cannot set nice value\n");
+				fwarning("cannot set nice value\n");
 				errno = 0;
 			}
 		}
 
-		// run cmdline trough shell
-		if (cfg.command_line == NULL) {
-			// if the sandbox was started with --shell=none, it is possible we don't have a shell
-			// inside the sandbox
-			if (cfg.shell == NULL) {
-				cfg.shell = guess_shell();
-				if (!cfg.shell) {
-					fprintf(stderr, "Error: no POSIX shell found, please use --shell command line option\n");
-					exit(1);
-				}
-			}
-				
-			struct stat s;
-			if (stat(cfg.shell, &s) == -1)  {
-				fprintf(stderr, "Error: %s shell not found inside the sandbox\n", cfg.shell);
-				exit(1);
-			}
+		// set environment, add x11 display
+		env_defaults();
+		if (display) {
+			char *display_str;
+			if (asprintf(&display_str, ":%d", display) == -1)
+				errExit("asprintf");
+			setenv("DISPLAY", display_str, 1);
+			free(display_str);
+		}
 
+		if (cfg.command_line == NULL) {
+			assert(cfg.shell);
 			cfg.command_line = cfg.shell;
 			cfg.window_title = cfg.shell;
 		}
@@ -396,6 +396,3 @@ void join(pid_t pid, int argc, char **argv, int index) {
 	flush_stdin();
 	exit(0);
 }
-
-
-
