@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -39,12 +39,12 @@ int is_container(const char *str) {
 // returns 1 if we are running under LXC
 int check_namespace_virt(void) {
 	EUID_ASSERT();
-	
+
 	// check container environment variable
 	char *str = getenv("container");
 	if (str && is_container(str))
 		return 1;
-	
+
 	// check PID 1 container environment variable
 	EUID_ROOT();
 	FILE *fp = fopen("/proc/1/environ", "r");
@@ -62,7 +62,7 @@ int check_namespace_virt(void) {
 					break;
 			}
 			buf[i] = '\0';
-			
+
 			// check env var name
 			if (strncmp(buf, "container=", 10) == 0) {
 				// found it
@@ -74,10 +74,10 @@ int check_namespace_virt(void) {
 			}
 //			printf("i %d c %d, buf #%s#\n", i, c, buf);
 		}
-		
+
 		fclose(fp);
 	}
-		
+
 	EUID_USER();
 	return 0;
 }
@@ -104,7 +104,7 @@ int check_kernel_procs(void) {
 
 	// look at the first 10 processes
 	// if a kernel process is found, return 1
-	for (i = 1; i <= 10; i++) { 
+	for (i = 1; i <= 10; i++) {
 		struct stat s;
 		char *fname;
 		if (asprintf(&fname, "/proc/%d/comm", i) == -1)
@@ -113,20 +113,20 @@ int check_kernel_procs(void) {
 			free(fname);
 			continue;
 		}
-		
+
 		// open file
 		/* coverity[toctou] */
 		FILE *fp = fopen(fname, "r");
 		if (!fp) {
-			fprintf(stderr, "Warning: cannot open %s\n", fname);
+			fwarning("cannot open %s\n", fname);
 			free(fname);
 			continue;
 		}
-		
+
 		// read file
 		char buf[100];
 		if (fgets(buf, 10, fp) == NULL) {
-			fprintf(stderr, "Warning: cannot read %s\n", fname);
+			fwarning("cannot read %s\n", fname);
 			fclose(fp);
 			free(fname);
 			continue;
@@ -135,7 +135,7 @@ int check_kernel_procs(void) {
 		char *ptr;
 		if ((ptr = strchr(buf, '\n')) != NULL)
 			*ptr = '\0';
-		
+
 		// check process name against the kernel list
 		int j = 0;
 		while (kern_proc[j] != NULL) {
@@ -148,7 +148,7 @@ int check_kernel_procs(void) {
 			}
 			j++;
 		}
-		
+
 		fclose(fp);
 		free(fname);
 	}
@@ -165,84 +165,28 @@ void run_no_sandbox(int argc, char **argv) {
 	// process limited subset of options
 	int i;
 	for (i = 0; i < argc; i++) {
-		if (strcmp(argv[i], "--csh") == 0) {
-			if (arg_shell_none) {
-				fprintf(stderr, "Error: --shell=none was already specified.\n");
-				exit(1);
-			}
-			if (cfg.shell) {
-				fprintf(stderr, "Error: only one default user shell can be specified\n");
-				exit(1);
-			}
-			cfg.shell = "/bin/csh";
-		}
-		else if (strcmp(argv[i], "--zsh") == 0) {
-			if (arg_shell_none) {
-				fprintf(stderr, "Error: --shell=none was already specified.\n");
-				exit(1);
-			}
-			if (cfg.shell) {
-				fprintf(stderr, "Error: only one default user shell can be specified\n");
-				exit(1);
-			}
-			cfg.shell = "/bin/zsh";
-		}
-		else if (strcmp(argv[i], "--shell=none") == 0) {
-			arg_shell_none = 1;
-			if (cfg.shell) {
-				fprintf(stderr, "Error: a shell was already specified\n");
-				exit(1);
-			}
-		}
-		else if (strncmp(argv[i], "--shell=", 8) == 0) {
-			if (arg_shell_none) {
-				fprintf(stderr, "Error: --shell=none was already specified.\n");
-				exit(1);
-			}
-			invalid_filename(argv[i] + 8);
-
-			if (cfg.shell) {
-				fprintf(stderr, "Error: only one user shell can be specified\n");
-				exit(1);
-			}
-			cfg.shell = argv[i] + 8;
-
-			if (is_dir(cfg.shell) || strstr(cfg.shell, "..")) {
-				fprintf(stderr, "Error: invalid shell\n");
-				exit(1);
-			}
-
-			// access call checks as real UID/GID, not as effective UID/GID
-			if(cfg.chrootdir) {
-				char *shellpath;
-				if (asprintf(&shellpath, "%s%s", cfg.chrootdir, cfg.shell) == -1)
-					errExit("asprintf");
-				if (access(shellpath, R_OK)) {
-					fprintf(stderr, "Error: cannot access shell file in chroot\n");
-					exit(1);
-				}
-				free(shellpath);
-			} else if (access(cfg.shell, R_OK)) {
-				fprintf(stderr, "Error: cannot access shell file\n");
-				exit(1);
-			}
-		}
+		if (strcmp(argv[i], "--debug") == 0)
+			arg_debug = 1;
+ 		else if (strcmp(argv[i], "--csh") == 0 ||
+		    strcmp(argv[i], "--zsh") == 0 ||
+		    strcmp(argv[i], "--shell=none") == 0 ||
+		    strncmp(argv[i], "--shell=", 8) == 0)
+			fwarning("shell-related command line options are disregarded - using SHELL environment variable\n");
 	}
 
 	// use $SHELL to get shell used in sandbox
-	if (!arg_shell_none && !cfg.shell) {
-		char *shell =  getenv("SHELL");
-		if (access(shell, R_OK) == 0)
-			cfg.shell = shell;
-	}
+	char *shell =  getenv("SHELL");
+	if (shell && access(shell, R_OK) == 0)
+		cfg.shell = shell;
+
 	// guess shell otherwise
-	if (!arg_shell_none && !cfg.shell) {
+	if (!cfg.shell) {
 		cfg.shell = guess_shell();
 		if (arg_debug)
 			printf("Autoselecting %s as shell\n", cfg.shell);
 	}
-	if (!arg_shell_none && !cfg.shell) {
-		fprintf(stderr, "Error: unable to guess your shell, please set explicitly by using --shell option.\n");
+	if (!cfg.shell) {
+		fprintf(stderr, "Error: unable to guess your shell, please set SHELL environment variable\n");
 		exit(1);
 	}
 
@@ -265,14 +209,17 @@ void run_no_sandbox(int argc, char **argv) {
 			break;
 		}
 	}
+	// if shell is /usr/bin/firejail, replace it with /bin/bash
+	if (strcmp(cfg.shell, PATH_FIREJAIL) == 0) {
+		cfg.shell = "/bin/bash";
+		prog_index = 0;
+	}
 
-	if (!arg_shell_none) {
-		if (prog_index == 0) {
-			cfg.command_line = cfg.shell;
-			cfg.window_title = cfg.shell;
-		} else {
-			build_cmdline(&cfg.command_line, &cfg.window_title, argc, argv, prog_index);
-		}
+	if (prog_index == 0) {
+		cfg.command_line = cfg.shell;
+		cfg.window_title = cfg.shell;
+	} else {
+		build_cmdline(&cfg.command_line, &cfg.window_title, argc, argv, prog_index);
 	}
 
 	cfg.original_argv = argv;
@@ -283,9 +230,10 @@ void run_no_sandbox(int argc, char **argv) {
 		command = cfg.shell;
 	else
 		command = argv[prog_index];
-	if (!arg_quiet)
-		fprintf(stderr, "Warning: an existing sandbox was detected. "
-			"%s will run without any additional sandboxing features\n", command);
+	fwarning("an existing sandbox was detected. "
+		"%s will run without any additional sandboxing features\n", command);
+
+	arg_quiet = 1;
 
 	start_application();
 }
