@@ -1,5 +1,5 @@
-    /*
- * Copyright (C) 2014-2016 Firejail Authors
+/*
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -27,17 +27,17 @@
 static void disable_file(const char *path, const char *file) {
 	assert(file);
 	assert(path);
-	
+
 	struct stat s;
 	char *fname;
 	if (asprintf(&fname, "%s/%s", path, file) == -1)
 		errExit("asprintf");
 	if (stat(fname, &s) == -1)
 		goto doexit;
-		
+
 	if (arg_debug)
 		printf("Disable%s\n", fname);
-		
+
 	if (S_ISDIR(s.st_mode)) {
 		if (mount(RUN_RO_DIR, fname, "none", MS_BIND, "mode=400,gid=0") < 0)
 			errExit("disable file");
@@ -71,7 +71,7 @@ void pulseaudio_disable(void) {
 		errExit("asprintf");
 	disable_file(path, "pulse/native");
 	free(path);
-		
+
 
 
 	// blacklist any pulse* file in /tmp directory
@@ -99,19 +99,16 @@ void pulseaudio_disable(void) {
 // disable shm in pulseaudio
 void pulseaudio_init(void) {
 	struct stat s;
-	
+
 	// do we have pulseaudio in the system?
 	if (stat("/etc/pulse/client.conf", &s) == -1)
 		return;
-	 
+
  	// create the new user pulseaudio directory
-	 fs_build_mnt_dir();
 	int rv = mkdir(RUN_PULSE_DIR, 0700);
 	(void) rv; // in --chroot mode the directory can already be there
-	if (chown(RUN_PULSE_DIR, getuid(), getgid()) < 0)
-		errExit("chown");
-	if (chmod(RUN_PULSE_DIR, 0700) < 0)
-		errExit("chmod");
+	if (set_perms(RUN_PULSE_DIR, getuid(), getgid(), 0700))
+		errExit("set_perms");
 
 	// create the new client.conf file
 	char *pulsecfg = NULL;
@@ -126,7 +123,6 @@ void pulseaudio_init(void) {
 	SET_PERMS_STREAM(fp, getuid(), getgid(), 0644);
 	fclose(fp);
 
-
 	// create ~/.config/pulse directory if not present
 	char *dir1;
 	if (asprintf(&dir1, "%s/.config", cfg.homedir) == -1)
@@ -138,12 +134,15 @@ void pulseaudio_init(void) {
 		if (child == 0) {
 			// drop privileges
 			drop_privs(0);
-	
+
 			int rv = mkdir(dir1, 0755);
 			if (rv == 0) {
 				if (set_perms(dir1, getuid(), getgid(), 0755))
 					{;} // do nothing
 			}
+#ifdef HAVE_GCOV
+			__gcov_flush();
+#endif
 			_exit(0);
 		}
 		// wait for the child to finish
@@ -157,7 +156,7 @@ void pulseaudio_init(void) {
 		}
 	}
 	free(dir1);
-	
+
 	if (asprintf(&dir1, "%s/.config/pulse", cfg.homedir) == -1)
 		errExit("asprintf");
 	if (stat(dir1, &s) == -1) {
@@ -167,12 +166,15 @@ void pulseaudio_init(void) {
 		if (child == 0) {
 			// drop privileges
 			drop_privs(0);
-	
+
 			int rv = mkdir(dir1, 0700);
 			if (rv == 0) {
 				if (set_perms(dir1, getuid(), getgid(), 0700))
 					{;} // do nothing
 			}
+#ifdef HAVE_GCOV
+			__gcov_flush();
+#endif
 			_exit(0);
 		}
 		// wait for the child to finish
@@ -186,13 +188,15 @@ void pulseaudio_init(void) {
 		}
 	}
 	free(dir1);
-	
+
+
 	// if we have ~/.config/pulse mount the new directory, else set environment variable
 	char *homeusercfg;
 	if (asprintf(&homeusercfg, "%s/.config/pulse", cfg.homedir) == -1)
 		errExit("asprintf");
 	if (stat(homeusercfg, &s) == 0) {
-		if (mount(RUN_PULSE_DIR, homeusercfg, "none", MS_BIND, NULL) < 0)
+		if (mount(RUN_PULSE_DIR, homeusercfg, "none", MS_BIND, NULL) < 0 ||
+		    mount(NULL, homeusercfg, NULL, MS_NOEXEC|MS_NODEV|MS_NOSUID|MS_BIND|MS_REMOUNT, NULL) < 0)
 			errExit("mount pulseaudio");
 		fs_logger2("tmpfs", homeusercfg);
 	}
@@ -201,7 +205,7 @@ void pulseaudio_init(void) {
 		if (setenv("PULSE_CLIENTCONFIG", pulsecfg, 1) < 0)
 			errExit("setenv");
 	}
-		
+
 	free(pulsecfg);
 	free(homeusercfg);
 }
