@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 Firejail Authors
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -17,7 +17,7 @@
  * with this program; if not, write to the Free Software Foundation, Inc.,
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
-	
+
 #include "firejail.h"
 #include <sys/mount.h>
 #include <sys/wait.h>
@@ -43,6 +43,12 @@
 #ifdef HAVE_APPARMOR
 #include <sys/apparmor.h>
 #endif
+#include <syscall.h>
+
+
+#ifdef HAVE_SECCOMP
+int enforce_seccomp = 0;
+#endif
 
 
 static int monitored_pid = 0;
@@ -65,7 +71,7 @@ static void sandbox_handler(int sig){
 			FILE *fp = fopen(monfile, "r");
 			if (!fp)
 				break;
-				
+
 			char c;
 			size_t count = fread(&c, 1, 1, fp);
 			fclose(fp);
@@ -78,7 +84,7 @@ static void sandbox_handler(int sig){
 			monsec--;
 		}
 		free(monfile);
-		
+
 	}
 
 
@@ -99,6 +105,11 @@ static void set_caps(void) {
 		caps_keep_list(arg_caps_list);
 	else if (arg_caps_default_filter)
 		caps_default_filter();
+	
+	// drop discretionary access control capabilities for root sandboxes
+	// if caps.keep, the user has to set it manually in the list
+	if (!arg_caps_keep)
+		caps_drop_dac_override();
 }
 
 void save_nogroups(void) {
@@ -115,14 +126,14 @@ void save_nogroups(void) {
 		fprintf(stderr, "Error: cannot save nogroups state\n");
 		exit(1);
 	}
-	
+
 }
 
 static void sandbox_if_up(Bridge *br) {
 	assert(br);
 	if (!br->configured)
 		return;
-		
+
 	char *dev = br->devsandbox;
 	net_if_up(dev);
 
@@ -132,13 +143,13 @@ static void sandbox_if_up(Bridge *br) {
 			fprintf(stderr, "Error: %d.%d.%d.%d is interface %s address.\n", PRINT_IP(br->ipsandbox), br->dev);
 			exit(1);
 		}
-		
+
 		// just assign the address
 		assert(br->ipsandbox);
 		if (arg_debug)
 			printf("Configuring %d.%d.%d.%d address on interface %s\n", PRINT_IP(br->ipsandbox), dev);
-		net_if_ip(dev, br->ipsandbox, br->mask, br->mtu);
-		net_if_up(dev);
+		net_config_interface(dev, br->ipsandbox, br->mask, br->mtu);
+		arp_announce(dev, br);
 	}
 	else if (br->arg_ip_none == 0 && br->macvlan == 1) {
 		// reassign the macvlan address
@@ -150,20 +161,20 @@ static void sandbox_if_up(Bridge *br) {
 				fprintf(stderr, "Error: %d.%d.%d.%d is interface %s address.\n", PRINT_IP(br->ipsandbox), br->dev);
 				exit(1);
 			}
-			
-			uint32_t rv = arp_check(dev, br->ipsandbox, br->ip);
+
+			uint32_t rv = arp_check(dev, br->ipsandbox);
 			if (rv) {
 				fprintf(stderr, "Error: the address %d.%d.%d.%d is already in use.\n", PRINT_IP(br->ipsandbox));
 				exit(1);
 			}
 		}
-			
+
 		if (arg_debug)
 			printf("Configuring %d.%d.%d.%d address on interface %s\n", PRINT_IP(br->ipsandbox), dev);
-		net_if_ip(dev, br->ipsandbox, br->mask, br->mtu);
-		net_if_up(dev);
+		net_config_interface(dev, br->ipsandbox, br->mask, br->mtu);
+		arp_announce(dev, br);
 	}
-	
+
 	if (br->ip6sandbox)
 		 net_if_ip6(dev, br->ip6sandbox);
 }
@@ -173,14 +184,14 @@ static void chk_chroot(void) {
 	char *mycont = getenv("container");
 	if (mycont)
 		return;
-	
+
 	// check if this is a regular chroot
 	struct stat s;
 	if (stat("/", &s) == 0) {
 		if (s.st_ino != 2)
 			return;
 	}
-	
+
 	fprintf(stderr, "Error: cannot mount filesystem as slave\n");
 	exit(1);
 }
@@ -210,6 +221,11 @@ static int monitor_application(pid_t app_pid) {
 		while(rv != monitored_pid);
 		if (arg_debug)
 			printf("Sandbox monitor: waitpid %u retval %d status %d\n", monitored_pid, rv, status);
+		if (rv == -1) { // we can get here if we have processes joining the sandbox (ECHILD)
+			if (arg_debug)
+				perror("waitpid");
+			sleep(1);
+		}
 
 		// if /proc is not remounted, we cannot check /proc directory,
 		// for now we just get out of here
@@ -235,15 +251,18 @@ static int monitor_application(pid_t app_pid) {
 				continue;
 			if (pid == 1)
 				continue;
-			
+
 			// todo: make this generic
 			// Dillo browser leaves a dpid process running, we need to shut it down
+			int found = 0;
 			if (strcmp(cfg.command_name, "dillo") == 0) {
 				char *pidname = pid_proc_comm(pid);
 				if (pidname && strcmp(pidname, "dpid") == 0)
-					break;
+					found = 1;
 				free(pidname);
 			}
+			if (found)
+				break;
 
 			monitored_pid = pid;
 			break;
@@ -256,47 +275,82 @@ static int monitor_application(pid_t app_pid) {
 
 	// return the latest exit status.
 	return status;
-
-#if 0
-// todo: find a way to shut down interfaces before closing the namespace
-// the problem is we don't have enough privileges to shutdown interfaces in this moment
-	// shut down bridge/macvlan interfaces
-	if (any_bridge_configured()) {
-		
-		if (cfg.bridge0.configured) {
-			printf("Shutting down %s\n", cfg.bridge0.devsandbox);
-			net_if_down( cfg.bridge0.devsandbox);
-		}
-		if (cfg.bridge1.configured) {
-			printf("Shutting down %s\n", cfg.bridge1.devsandbox);
-			net_if_down( cfg.bridge1.devsandbox);
-		}
-		if (cfg.bridge2.configured) {
-			printf("Shutting down %s\n", cfg.bridge2.devsandbox);
-			net_if_down( cfg.bridge2.devsandbox);
-		}
-		if (cfg.bridge3.configured) {
-			printf("Shutting down %s\n", cfg.bridge3.devsandbox);
-			net_if_down( cfg.bridge3.devsandbox);
-		}
-		usleep(20000);	// 20 ms sleep
-	}	
-#endif	
 }
 
-void start_audit(void) {
-	char *audit_prog;
-	if (asprintf(&audit_prog, "%s/firejail/faudit", LIBDIR) == -1)
-		errExit("asprintf");
-	execl(audit_prog, audit_prog, NULL);
-	perror("execl");
-	exit(1);
+static void print_time(void) {
+	if (start_timestamp) {
+		unsigned long long end_timestamp = getticks();
+		// measure 1 ms
+		usleep(1000);
+		unsigned long long onems = getticks() - end_timestamp;
+		if (onems) {
+			printf("Child process initialized in %.02f ms\n",
+				(float) (end_timestamp - start_timestamp) / (float) onems);
+			return;
+		}
+	}
+
+	printf("Child process initialized\n");
+}
+
+
+// check execute permissions for the program
+// this is done typically by the shell
+// we are here because of --shell=none
+// we duplicate execvp functionality (man execvp):
+//	[...] if  the  specified
+//	filename  does  not contain a slash (/) character. The file is sought
+//	in the colon-separated list of directory pathnames  specified  in  the
+//	PATH  environment  variable.
+static int ok_to_run(const char *program) {
+	if (strstr(program, "/")) {
+		if (access(program, X_OK) == 0) // it will also dereference symlinks
+			return 1;
+	}
+	else { // search $PATH
+		char *path1 = getenv("PATH");
+		if (path1) {
+			if (arg_debug)
+				printf("Searching $PATH for %s\n", program);
+			char *path2 = strdup(path1);
+			if (!path2)
+				errExit("strdup");
+
+			// use path2 to count the entries
+			char *ptr = strtok(path2, ":");
+			while (ptr) {
+				char *fname;
+
+				if (asprintf(&fname, "%s/%s", ptr, program) == -1)
+					errExit("asprintf");
+				if (arg_debug)
+					printf("trying #%s#\n", fname);
+
+				struct stat s;
+				int rv = stat(fname, &s);
+				if (rv == 0) {
+					if (access(fname, X_OK) == 0) {
+						free(path2);
+						free(fname);
+						return 1;
+					}
+					else
+						fprintf(stderr, "Error: execute permission denied for %s\n", fname);
+
+					free(fname);
+					break;
+				}
+
+				free(fname);
+				ptr = strtok(NULL, ":");
+			}
+			free(path2);
+		}
+	}
+	return 0;
 }
 
 void start_application(void) {
-//if (setsid() == -1)
-//errExit("setsid");
-
 	// set environment
 	env_defaults();
 	env_apply();
@@ -310,11 +364,12 @@ void start_application(void) {
 	//****************************************
 	if (arg_audit) {
 		assert(arg_audit_prog);
-		struct stat s;
-		if (stat(arg_audit_prog, &s) != 0) {
-			fprintf(stderr, "Error: cannot find the audit program\n");
-			exit(1);
-		}
+#ifdef HAVE_GCOV
+		__gcov_dump();
+#endif
+#ifdef HAVE_SECCOMP
+		seccomp_install_filters();
+#endif
 		execl(arg_audit_prog, arg_audit_prog, NULL);
 	}
 	//****************************************
@@ -336,9 +391,19 @@ void start_application(void) {
 		}
 
 		if (!arg_command && !arg_quiet)
-			printf("Child process initialized\n");
+			print_time();
 
-		execvp(cfg.original_argv[cfg.original_program_index], &cfg.original_argv[cfg.original_program_index]);
+		int rv = ok_to_run(cfg.original_argv[cfg.original_program_index]);
+#ifdef HAVE_GCOV
+		__gcov_dump();
+#endif
+#ifdef HAVE_SECCOMP
+		seccomp_install_filters();
+#endif
+		if (rv)
+			execvp(cfg.original_argv[cfg.original_program_index], &cfg.original_argv[cfg.original_program_index]);
+		else
+			fprintf(stderr, "Error: no suitable %s executable found\n", cfg.original_argv[cfg.original_program_index]);
 		exit(1);
 	}
 	//****************************************
@@ -384,7 +449,13 @@ void start_application(void) {
 		}
 
 		if (!arg_command && !arg_quiet)
-			printf("Child process initialized\n");
+			print_time();
+#ifdef HAVE_GCOV
+		__gcov_dump();
+#endif
+#ifdef HAVE_SECCOMP
+		seccomp_install_filters();
+#endif
 		execvp(arg[0], arg);
 	}
 
@@ -396,6 +467,9 @@ static void enforce_filters(void) {
 	// force default seccomp inside the chroot, no keep or drop list
 	// the list build on top of the default drop list is kept intact
 	arg_seccomp = 1;
+#ifdef HAVE_SECCOMP
+	enforce_seccomp = 1;
+#endif
 	if (cfg.seccomp_list_drop) {
 		free(cfg.seccomp_list_drop);
 		cfg.seccomp_list_drop = NULL;
@@ -404,12 +478,12 @@ static void enforce_filters(void) {
 		free(cfg.seccomp_list_keep);
 		cfg.seccomp_list_keep = NULL;
 	}
-	
+
 	// disable all capabilities
 	if (arg_caps_default_filter || arg_caps_list)
-		fprintf(stderr, "Warning: all capabilities disabled for a regular user in chroot\n");
+		fwarning("all capabilities disabled for a regular user in chroot\n");
 	arg_caps_drop_all = 1;
-	
+
 	// drop all supplementary groups; /etc/group file inside chroot
 	// is controlled by a regular usr
 	arg_nogroups = 1;
@@ -423,12 +497,12 @@ int sandbox(void* sandbox_arg) {
 
 	pid_t child_pid = getpid();
 	if (arg_debug)
-		printf("Initializing child process\n");	
+		printf("Initializing child process\n");
 
  	// close each end of the unused pipes
  	close(parent_to_child_fds[1]);
  	close(child_to_parent_fds[0]);
- 
+
  	// wait for parent to do base setup
  	wait_for_other(parent_to_child_fds[0]);
 
@@ -451,8 +525,9 @@ int sandbox(void* sandbox_arg) {
 	if (mount(NULL, "/", NULL, MS_SLAVE | MS_REC, NULL) < 0) {
 		chk_chroot();
 	}
-	
-	
+	// ... and mount a tmpfs on top of /run/firejail/mnt directory
+	preproc_mount_mnt_dir();
+
 	//****************************
 	// log sandbox data
 	//****************************
@@ -461,14 +536,14 @@ int sandbox(void* sandbox_arg) {
 	fs_logger2int("sandbox pid:", (int) sandbox_pid);
 	if (cfg.chrootdir)
 		fs_logger("sandbox filesystem: chroot");
-	else if (arg_overlay)	
+	else if (arg_overlay)
 		fs_logger("sandbox filesystem: overlay");
 	else
 		fs_logger("sandbox filesystem: local");
 	fs_logger("install mount namespace");
-	
+
 	//****************************
-	// netfilter etc.
+	// netfilter
 	//****************************
 	if (arg_netfilter && any_bridge_configured()) { // assuming by default the client filter
 		netfilter(arg_netfilter_file);
@@ -477,194 +552,6 @@ int sandbox(void* sandbox_arg) {
 		netfilter6(arg_netfilter6_file);
 	}
 
-	// load IBUS env variables
-	if (arg_nonetwork || any_bridge_configured() || any_interface_configured()) {
-		// do nothing - there are problems with ibus version 1.5.11
-	}
-	else
-		env_ibus_load();
-	
-	// grab a copy of cp command
-	fs_build_cp_command();
-	
-	// trace pre-install
-	if (arg_trace || arg_tracelog || mask_x11_abstract_socket)
-		fs_trace_preload();
-
-	//****************************
-	// configure filesystem
-	//****************************
-#ifdef HAVE_SECCOMP
-	int enforce_seccomp = 0;
-#endif
-#ifdef HAVE_CHROOT		
-	if (cfg.chrootdir) {
-		fs_chroot(cfg.chrootdir);
-		
-		// force caps and seccomp if not started as root
-		if (getuid() != 0) {
-			enforce_filters();
-#ifdef HAVE_SECCOMP
-			enforce_seccomp = 1;
-#endif
-		}
-		else
-			arg_seccomp = 1;
-						
-		//****************************
-		// trace pre-install, this time inside chroot
-		//****************************
-		if (arg_trace || arg_tracelog || mask_x11_abstract_socket)
-			fs_trace_preload();
-	}
-	else 
-#endif		
-#ifdef HAVE_OVERLAYFS
-	if (arg_overlay)	{
-		fs_overlayfs();
-		// force caps and seccomp if not started as root
-		if (getuid() != 0) {
-			enforce_filters();
-#ifdef HAVE_SECCOMP
-			enforce_seccomp = 1;
-#endif
-		}
-		else
-			arg_seccomp = 1;
-	}
-	else
-#endif
-		fs_basic_fs();
-	
-	//****************************
-	// set hostname in /etc/hostname
-	//****************************
-	if (cfg.hostname) {
-		fs_hostname(cfg.hostname);
-	}
-	
-	//****************************
-	// private mode
-	//****************************
-	if (arg_private) {
-		if (cfg.home_private) {	// --private=
-			if (cfg.chrootdir)
-				fprintf(stderr, "Warning: private=directory feature is disabled in chroot\n");
-			else if (arg_overlay)
-				fprintf(stderr, "Warning: private=directory feature is disabled in overlay\n");
-			else
-				fs_private_homedir();
-		}
-		else if (cfg.home_private_keep) { // --private-home=
-			if (cfg.chrootdir)
-				fprintf(stderr, "Warning: private-home= feature is disabled in chroot\n");
-			else if (arg_overlay)
-				fprintf(stderr, "Warning: private-home= feature is disabled in overlay\n");
-			else
-				fs_private_home_list();
-		}
-		else // --private
-			fs_private();
-	}
-
-	if (arg_private_dev) {
-		if (cfg.chrootdir)
-			fprintf(stderr, "Warning: private-dev feature is disabled in chroot\n");
-		else if (arg_overlay)
-			fprintf(stderr, "Warning: private-dev feature is disabled in overlay\n");
-		else
-			fs_private_dev();
-	}
-		
-	if (arg_private_etc) {
-		if (cfg.chrootdir)
-			fprintf(stderr, "Warning: private-etc feature is disabled in chroot\n");
-		else if (arg_overlay)
-			fprintf(stderr, "Warning: private-etc feature is disabled in overlay\n");
-		else {
-			fs_private_etc_list();
-			// create /etc/ld.so.preload file again
-			if (arg_trace || arg_tracelog || mask_x11_abstract_socket)
-				fs_trace_preload();
-		}
-	}
-	
-	if (arg_private_bin) {
-		if (cfg.chrootdir)
-			fprintf(stderr, "Warning: private-bin feature is disabled in chroot\n");
-		else if (arg_overlay)
-			fprintf(stderr, "Warning: private-bin feature is disabled in overlay\n");
-		else {
-			// for --x11=xorg we need to add xauth command
-			if (arg_x11_xorg) {
-				EUID_USER();
-				char *tmp;
-				if (asprintf(&tmp, "%s,xauth", cfg.bin_private_keep) == -1)
-					errExit("asprintf");
-				cfg.bin_private_keep = tmp;
-				fs_check_bin_list();
-				EUID_ROOT();
-			}
-			fs_private_bin_list();
-		}
-	}
-	
-	if (arg_private_tmp) {
-		if (cfg.chrootdir)
-			fprintf(stderr, "Warning: private-tmp feature is disabled in chroot\n");
-		else if (arg_overlay)
-			fprintf(stderr, "Warning: private-tmp feature is disabled in overlay\n");
-		else {
-			// private-tmp is implemented as a whitelist
-			EUID_USER();
-			profile_add("whitelist /tmp/.X11-unix");
-			EUID_ROOT();
-//			fs_private_tmp();
-		}
-	}
-	
-	//****************************
-	// update /proc, /sys, /dev, /boot directorymy
-	//****************************
-	if (checkcfg(CFG_REMOUNT_PROC_SYS))
-		fs_proc_sys_dev_boot();
-	
-	//****************************
-	// apply the profile file
-	//****************************
-	// apply all whitelist commands ... 
-	if (cfg.chrootdir)
-		fprintf(stderr, "Warning: whitelist feature is disabled in chroot\n");
-	else if (arg_overlay)
-		fprintf(stderr, "Warning: whitelist feature is disabled in overlay\n");
-	else
-		fs_whitelist();
-	
-	// ... followed by blacklist commands
-	fs_blacklist();
-	
-	//****************************
-	// install trace
-	//****************************
-	if (arg_trace || arg_tracelog || mask_x11_abstract_socket)
-		fs_trace();
-		
-	//****************************
-	// nosound/no3d and fix for pulseaudio 7.0
-	//****************************
-	if (arg_nosound) {
-		// disable pulseaudio
-		pulseaudio_disable();
-
-		// disable /dev/snd
-		fs_dev_disable_sound();
-	}
-	else
-		pulseaudio_init();
-	
-	if (arg_no3d)
-		fs_dev_disable_3d();
-	
 	//****************************
 	// networking
 	//****************************
@@ -674,57 +561,60 @@ int sandbox(void* sandbox_arg) {
 		if (arg_debug)
 			printf("Network namespace enabled, only loopback interface available\n");
 	}
+	else if (arg_netns) {
+		netns(arg_netns);
+		if (arg_debug)
+			printf("Network namespace '%s' activated\n", arg_netns);
+	}
 	else if (any_bridge_configured() || any_interface_configured()) {
 		// configure lo and eth0...eth3
 		net_if_up("lo");
-		
+
 		if (mac_not_zero(cfg.bridge0.macsandbox))
 			net_config_mac(cfg.bridge0.devsandbox, cfg.bridge0.macsandbox);
 		sandbox_if_up(&cfg.bridge0);
-		
+
 		if (mac_not_zero(cfg.bridge1.macsandbox))
 			net_config_mac(cfg.bridge1.devsandbox, cfg.bridge1.macsandbox);
 		sandbox_if_up(&cfg.bridge1);
-		
+
 		if (mac_not_zero(cfg.bridge2.macsandbox))
 			net_config_mac(cfg.bridge2.devsandbox, cfg.bridge2.macsandbox);
 		sandbox_if_up(&cfg.bridge2);
-		
+
 		if (mac_not_zero(cfg.bridge3.macsandbox))
 			net_config_mac(cfg.bridge3.devsandbox, cfg.bridge3.macsandbox);
 		sandbox_if_up(&cfg.bridge3);
-		
-		// enable interfaces
+
+
+		// moving an interface in a namespace using --interface will reset the interface configuration;
+		// we need to put the configuration back
 		if (cfg.interface0.configured && cfg.interface0.ip) {
 			if (arg_debug)
 				printf("Configuring %d.%d.%d.%d address on interface %s\n", PRINT_IP(cfg.interface0.ip), cfg.interface0.dev);
-			net_if_ip(cfg.interface0.dev, cfg.interface0.ip, cfg.interface0.mask, cfg.interface0.mtu);
-			net_if_up(cfg.interface0.dev);
-		}			
+			net_config_interface(cfg.interface0.dev, cfg.interface0.ip, cfg.interface0.mask, cfg.interface0.mtu);
+		}
 		if (cfg.interface1.configured && cfg.interface1.ip) {
 			if (arg_debug)
 				printf("Configuring %d.%d.%d.%d address on interface %s\n", PRINT_IP(cfg.interface1.ip), cfg.interface1.dev);
-			net_if_ip(cfg.interface1.dev, cfg.interface1.ip, cfg.interface1.mask, cfg.interface1.mtu);
-			net_if_up(cfg.interface1.dev);
-		}			
+			net_config_interface(cfg.interface1.dev, cfg.interface1.ip, cfg.interface1.mask, cfg.interface1.mtu);
+		}
 		if (cfg.interface2.configured && cfg.interface2.ip) {
 			if (arg_debug)
 				printf("Configuring %d.%d.%d.%d address on interface %s\n", PRINT_IP(cfg.interface2.ip), cfg.interface2.dev);
-			net_if_ip(cfg.interface2.dev, cfg.interface2.ip, cfg.interface2.mask, cfg.interface2.mtu);
-			net_if_up(cfg.interface2.dev);
-		}			
+			net_config_interface(cfg.interface2.dev, cfg.interface2.ip, cfg.interface2.mask, cfg.interface2.mtu);
+		}
 		if (cfg.interface3.configured && cfg.interface3.ip) {
 			if (arg_debug)
 				printf("Configuring %d.%d.%d.%d address on interface %s\n", PRINT_IP(cfg.interface3.ip), cfg.interface3.dev);
-			net_if_ip(cfg.interface3.dev, cfg.interface3.ip, cfg.interface3.mask, cfg.interface3.mtu);
-			net_if_up(cfg.interface3.dev);
-		}			
-			
+			net_config_interface(cfg.interface3.dev, cfg.interface3.ip, cfg.interface3.mask, cfg.interface3.mtu);
+		}
+
 		// add a default route
 		if (cfg.defaultgw) {
 			// set the default route
 			if (net_add_route(0, 0, cfg.defaultgw)) {
-				fprintf(stderr, "Warning: cannot configure default route\n");
+				fwarning("cannot configure default route\n");
 				gw_cfg_failed = 1;
 			}
 		}
@@ -732,18 +622,19 @@ int sandbox(void* sandbox_arg) {
 		if (arg_debug)
 			printf("Network namespace enabled\n");
 	}
-	
-	// if any dns server is configured, it is time to set it now
-	fs_resolvconf();
-	fs_logger_print();
-	fs_logger_change_owner();
 
 	// print network configuration
 	if (!arg_quiet) {
 		if (any_bridge_configured() || any_interface_configured() || cfg.defaultgw || cfg.dns1) {
 			printf("\n");
-			if (any_bridge_configured() || any_interface_configured())
-				net_ifprint();
+			if (any_bridge_configured() || any_interface_configured()) {
+//				net_ifprint();
+				if (arg_scan)
+					sbox_run(SBOX_ROOT | SBOX_CAPS_NETWORK | SBOX_SECCOMP, 3, PATH_FNET, "printif", "scan");
+				else
+					sbox_run(SBOX_ROOT | SBOX_CAPS_NETWORK | SBOX_SECCOMP, 2, PATH_FNET, "printif", "scan");
+
+			}
 			if (cfg.defaultgw != 0) {
 				if (gw_cfg_failed)
 					printf("Default gateway configuration failed\n");
@@ -759,8 +650,272 @@ int sandbox(void* sandbox_arg) {
 			printf("\n");
 		}
 	}
-	
-	fs_delete_cp_command();
+
+	// load IBUS env variables
+	if (arg_nonetwork || any_bridge_configured() || any_interface_configured()) {
+		// do nothing - there are problems with ibus version 1.5.11
+	}
+	else
+		env_ibus_load();
+
+	//****************************
+	// fs pre-processing:
+	//  - build seccomp filters
+	//  - create an empty /etc/ld.so.preload
+	//****************************
+#ifdef HAVE_SECCOMP
+	if (cfg.protocol) {
+		if (arg_debug)
+			printf("Build protocol filter: %s\n", cfg.protocol);
+
+		// build the seccomp filter as a regular user
+		int rv = sbox_run(SBOX_USER | SBOX_CAPS_NONE | SBOX_SECCOMP, 5,
+			PATH_FSECCOMP, "protocol", "build", cfg.protocol, RUN_SECCOMP_PROTOCOL);
+		if (rv)
+			exit(rv);
+	}
+	if (arg_seccomp && (cfg.seccomp_list || cfg.seccomp_list_drop || cfg.seccomp_list_keep))
+		arg_seccomp_postexec = 1;
+#endif
+
+	// need ld.so.preload if tracing or seccomp with any non-default lists
+	bool need_preload = arg_trace || arg_tracelog || arg_seccomp_postexec;
+
+	// trace pre-install
+	if (need_preload)
+		fs_trace_preload();
+
+	// store hosts file
+	if (cfg.hosts_file)
+		fs_store_hosts_file();
+
+	//****************************
+	// configure filesystem
+	//****************************
+	if (arg_appimage)
+		enforce_filters();
+
+#ifdef HAVE_CHROOT
+	if (cfg.chrootdir) {
+		fs_chroot(cfg.chrootdir);
+
+		// force caps and seccomp if not started as root
+		if (getuid() != 0)
+			enforce_filters();
+		else
+			arg_seccomp = 1;
+
+		//****************************
+		// trace pre-install, this time inside chroot
+		//****************************
+		if (need_preload)
+			fs_trace_preload();
+	}
+	else
+#endif
+#ifdef HAVE_OVERLAYFS
+	if (arg_overlay)	{
+		fs_overlayfs();
+		// force caps and seccomp if not started as root
+		if (getuid() != 0)
+			enforce_filters();
+		else
+			arg_seccomp = 1;
+	}
+	else
+#endif
+		fs_basic_fs();
+
+	//****************************
+	// private mode
+	//****************************
+	if (arg_private) {
+		if (cfg.home_private) {	// --private=
+			if (cfg.chrootdir)
+				fwarning("private=directory feature is disabled in chroot\n");
+			else if (arg_overlay)
+				fwarning("private=directory feature is disabled in overlay\n");
+			else
+				fs_private_homedir();
+		}
+		else if (cfg.home_private_keep) { // --private-home=
+			if (cfg.chrootdir)
+				fwarning("private-home= feature is disabled in chroot\n");
+			else if (arg_overlay)
+				fwarning("private-home= feature is disabled in overlay\n");
+			else
+				fs_private_home_list();
+		}
+		else // --private
+			fs_private();
+	}
+
+	if (arg_private_dev) {
+		if (cfg.chrootdir)
+			fwarning("private-dev feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-dev feature is disabled in overlay\n");
+		else
+			fs_private_dev();
+	}
+
+	if (arg_private_etc) {
+		if (cfg.chrootdir)
+			fwarning("private-etc feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-etc feature is disabled in overlay\n");
+		else {
+			fs_private_dir_list("/etc", RUN_ETC_DIR, cfg.etc_private_keep);
+			// create /etc/ld.so.preload file again
+			if (need_preload)
+				fs_trace_preload();
+		}
+	}
+
+	if (arg_private_opt) {
+		if (cfg.chrootdir)
+			fwarning("private-opt feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-opt feature is disabled in overlay\n");
+		else {
+			fs_private_dir_list("/opt", RUN_OPT_DIR, cfg.opt_private_keep);
+		}
+	}
+
+	if (arg_private_srv) {
+		if (cfg.chrootdir)
+			fwarning("private-srv feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-srv feature is disabled in overlay\n");
+		else {
+			fs_private_dir_list("/srv", RUN_SRV_DIR, cfg.srv_private_keep);
+		}
+	}
+
+	if (arg_private_bin) {
+		if (cfg.chrootdir)
+			fwarning("private-bin feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-bin feature is disabled in overlay\n");
+		else {
+			// for --x11=xorg we need to add xauth command
+			if (arg_x11_xorg) {
+				EUID_USER();
+				char *tmp;
+				if (asprintf(&tmp, "%s,xauth", cfg.bin_private_keep) == -1)
+					errExit("asprintf");
+				cfg.bin_private_keep = tmp;
+				EUID_ROOT();
+			}
+			fs_private_bin_list();
+		}
+	}
+
+	if (arg_private_lib) {
+		if (cfg.chrootdir)
+			fwarning("private-lib feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-lib feature is disabled in overlay\n");
+		else {
+			fs_private_lib();
+		}
+	}
+
+	if (arg_private_tmp) {
+		if (cfg.chrootdir)
+			fwarning("private-tmp feature is disabled in chroot\n");
+		else if (arg_overlay)
+			fwarning("private-tmp feature is disabled in overlay\n");
+		else {
+			// private-tmp is implemented as a whitelist
+			EUID_USER();
+			fs_private_tmp();
+			EUID_ROOT();
+		}
+	}
+
+	//****************************
+	// hosts and hostname
+	//****************************
+	if (cfg.hostname)
+		fs_hostname(cfg.hostname);
+
+	if (cfg.hosts_file)
+		fs_mount_hosts_file();
+
+	//****************************
+	// /etc overrides from the network namespace
+	//****************************
+	if (arg_netns)
+		netns_mounts(arg_netns);
+
+	//****************************
+	// update /proc, /sys, /dev, /boot directory
+	//****************************
+	if (checkcfg(CFG_REMOUNT_PROC_SYS))
+		fs_proc_sys_dev_boot();
+
+	//****************************
+	// handle /mnt and /media
+	//****************************
+	if (arg_disable_mnt || checkcfg(CFG_DISABLE_MNT))
+		fs_mnt();
+
+	//****************************
+	// apply the profile file
+	//****************************
+	// apply all whitelist commands ...
+	if (cfg.chrootdir)
+		fwarning("whitelist feature is disabled in chroot\n");
+	else if (arg_overlay)
+		fwarning("whitelist feature is disabled in overlay\n");
+	else
+		fs_whitelist();
+
+	// ... followed by blacklist commands
+	fs_blacklist(); // mkdir and mkfile are processed all over again
+
+	//****************************
+	// nosound/no3d/notv/novideo and fix for pulseaudio 7.0
+	//****************************
+	if (arg_nosound) {
+		// disable pulseaudio
+		pulseaudio_disable();
+
+		// disable /dev/snd
+		fs_dev_disable_sound();
+	}
+	else
+		pulseaudio_init();
+
+	if (arg_no3d)
+		fs_dev_disable_3d();
+
+	if (arg_notv)
+		fs_dev_disable_tv();
+
+	if (arg_nodvd)
+		fs_dev_disable_dvd();
+
+	if (arg_novideo)
+		fs_dev_disable_video();
+
+	//****************************
+	// install trace
+	//****************************
+	if (need_preload)
+		fs_trace();
+
+	//****************************
+	// set dns
+	//****************************
+	fs_resolvconf();
+
+	//****************************
+	// fs post-processing
+	//****************************
+	fs_logger_print();
+	fs_logger_change_owner();
 
 	//****************************
 	// set application environment
@@ -771,7 +926,7 @@ int sandbox(void* sandbox_arg) {
 		if (chdir(cfg.cwd) == 0)
 			cwd = 1;
 	}
-	
+
 	if (!cwd) {
 		if (chdir("/") < 0)
 			errExit("chdir");
@@ -784,62 +939,84 @@ int sandbox(void* sandbox_arg) {
 			}
 		}
 	}
-	
+	if (arg_debug) {
+		char *cpath = get_current_dir_name();
+		if (cpath) {
+			printf("Current directory: %s\n", cpath);
+			free(cpath);
+		}
+	}
+
+
 	// set nice
 	if (arg_nice) {
 		errno = 0;
 		int rv = nice(cfg.nice);
 		(void) rv;
 		if (errno) {
-			fprintf(stderr, "Warning: cannot set nice value\n");
+			fwarning("cannot set nice value\n");
 			errno = 0;
 		}
 	}
-	
+
 	// clean /tmp/.X11-unix sockets
 	fs_x11();
 	if (arg_x11_xorg)
 		x11_xorg();
-	
+
 	//****************************
 	// set security filters
 	//****************************
 	// set capabilities
-//	if (!arg_noroot)
-		set_caps();
+	set_caps();
 
 	// set rlimits
 	set_rlimits();
 
-	// set seccomp
-#ifdef HAVE_SECCOMP
-	// install protocol filter
-	if (cfg.protocol) {
-		protocol_filter();	// install filter	
-		protocol_filter_save();	// save filter in PROTOCOL_CFG
-	}
-
-	// if a keep list is available, disregard the drop list
-	if (arg_seccomp == 1) {
-		if (cfg.seccomp_list_keep)
-			seccomp_filter_keep();
-		else if (cfg.seccomp_list_errno)
-			seccomp_filter_errno(); 
-		else
-			seccomp_filter_drop(enforce_seccomp);
-	}
-#endif
-
 	// set cpu affinity
 	if (cfg.cpus) {
 		save_cpu(); // save cpu affinity mask to CPU_CFG file
 		set_cpu_affinity();
 	}
-	
+
 	// save cgroup in CGROUP_CFG file
 	if (cfg.cgroup)
 		save_cgroup();
 
+	// set seccomp
+#ifdef HAVE_SECCOMP
+	// install protocol filter
+#ifdef SYS_socket
+	if (cfg.protocol) {
+		if (arg_debug)
+			printf("Install protocol filter: %s\n", cfg.protocol);
+		seccomp_load(RUN_SECCOMP_PROTOCOL);	// install filter
+		protocol_filter_save();	// save filter in RUN_PROTOCOL_CFG
+	}
+#endif
+
+	// if a keep list is available, disregard the drop list
+	if (arg_seccomp == 1) {
+		if (cfg.seccomp_list_keep)
+			seccomp_filter_keep();
+		else
+			seccomp_filter_drop();
+	}
+
+	if (arg_debug) {
+		printf("\nSeccomp files:\n");
+		int rv = system("ls -l /run/firejail/mnt/seccomp*\n");
+		(void) rv;
+		printf("\n");
+	}
+
+	if (arg_memory_deny_write_execute) {
+		if (arg_debug)
+			printf("Install memory write&execute filter\n");
+		seccomp_load(RUN_SECCOMP_MDWX);	// install filter
+	}
+#endif
+
 	//****************************************
 	// drop privileges or create a new user namespace
 	//****************************************
@@ -847,19 +1024,19 @@ int sandbox(void* sandbox_arg) {
 	if (arg_noroot) {
 		int rv = unshare(CLONE_NEWUSER);
 		if (rv == -1) {
-			fprintf(stderr, "Warning: cannot create a new user namespace, going forward without it...\n");
+			fwarning("cannot create a new user namespace, going forward without it...\n");
 			drop_privs(arg_nogroups);
 			arg_noroot = 0;
 		}
 	}
 	else
 		drop_privs(arg_nogroups);
-	
+
 	// notify parent that new user namespace has been created so a proper
  	// UID/GID map can be setup
  	notify_other(child_to_parent_fds[1]);
  	close(child_to_parent_fds[1]);
- 
+
  	// wait for parent to finish setting up a proper UID/GID map
  	wait_for_other(parent_to_child_fds[0]);
  	close(parent_to_child_fds[0]);
@@ -871,15 +1048,15 @@ int sandbox(void* sandbox_arg) {
 			printf("noroot user namespace installed\n");
 		set_caps();
 	}
-	
+
 	//****************************************
 	// Set NO_NEW_PRIVS if desired
 	//****************************************
 	if (arg_nonewprivs) {
 		int no_new_privs = prctl(PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0);
 
-		if(no_new_privs != 0)
-			fprintf(stderr, "Warning: NO_NEW_PRIVS disabled, it requires a Linux kernel version 3.5 or newer.\n");
+		if(no_new_privs != 0 && !arg_quiet)
+			fwarning("NO_NEW_PRIVS disabled, it requires a Linux kernel version 3.5 or newer.\n");
 		else if (arg_debug)
 			printf("NO_NEW_PRIVS set\n");
 	}
@@ -896,15 +1073,14 @@ int sandbox(void* sandbox_arg) {
 		if (arg_apparmor) {
 			errno = 0;
 			if (aa_change_onexec("firejail-default")) {
-				fprintf(stderr, "Error: cannot confine the application using AppArmor.\n");
-				fprintf(stderr, "Maybe firejail-default AppArmor profile is not loaded into the kernel.\n");
-				fprintf(stderr, "As root, run \"aa-enforce firejail-default\" to load it.\n");
-				exit(1);
+				fwarning("Cannot confine the application using AppArmor.\n"
+					"Maybe firejail-default AppArmor profile is not loaded into the kernel.\n"
+					"As root, run \"aa-enforce firejail-default\" to load it.\n");
 			}
 			else if (arg_debug)
 				printf("AppArmor enabled\n");
 		}
-#endif		
+#endif
 		prctl(PR_SET_PDEATHSIG, SIGKILL, 0, 0, 0); // kill the child in case the parent died
 		start_application();	// start app
 	}
@@ -912,8 +1088,6 @@ int sandbox(void* sandbox_arg) {
 	int status = monitor_application(app_pid);	// monitor application
 	flush_stdin();
 
-
-
 	if (WIFEXITED(status)) {
 		// if we had a proper exit, return that exit status
 		return WEXITSTATUS(status);
