@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 netblue30 (netblue30@yahoo.com)
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -38,7 +38,6 @@ static inline void firemon_clrscr(void) {
 // firemon.c
 extern int arg_nowrap;
 int find_child(int id);
-void firemon_drop_privs(void);
 void firemon_sleep(int st);
 
 
@@ -55,25 +54,25 @@ void top(void);
 void list(void);
 
 // interface.c
-void interface(pid_t pid);
+void interface(pid_t pid, int print_procs);
 
 // arp.c
-void arp(pid_t pid);
+void arp(pid_t pid, int print_procs);
 
 // route.c
-void route(pid_t pid);
+void route(pid_t pid, int print_procs);
 
 // caps.c
-void caps(pid_t pid);
+void caps(pid_t pid, int print_procs);
 
 // seccomp.c
-void seccomp(pid_t pid);
+void seccomp(pid_t pid, int print_procs);
 
 // cpu.c
-void cpu(pid_t pid);
+void cpu(pid_t pid, int print_procs);
 
 // cgroup.c
-void cgroup(pid_t pid);
+void cgroup(pid_t pid, int print_procs);
 
 // tree.c
 void tree(pid_t pid);
@@ -82,6 +81,6 @@ void tree(pid_t pid);
 void netstats(void);
 
 // x11.c
-void x11(pid_t pid);
+void x11(pid_t pid, int print_procs);
 
 #endif
