@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2014-2016 netblue30 (netblue30@yahoo.com)
+ * Copyright (C) 2014-2017 Firejail Authors
  *
  * This file is part of firejail project
  *
@@ -37,31 +37,32 @@ int join_namespace(pid_t pid, char *type) {
 	char *path;
 	if (asprintf(&path, "/proc/%u/ns/%s", pid, type) == -1)
 		errExit("asprintf");
-	
+
 	int fd = open(path, O_RDONLY);
-	if (fd < 0) {
-		free(path);
-		fprintf(stderr, "Error: cannot open /proc/%u/ns/%s.\n", pid, type);
-		return -1;
-	}
+	if (fd < 0)
+		goto errout;
 
 	if (syscall(__NR_setns, fd, 0) < 0) {
-		free(path);
-		fprintf(stderr, "Error: cannot join namespace %s.\n", type);
 		close(fd);
-		return -1;
+		goto errout;
 	}
 
 	close(fd);
 	free(path);
 	return 0;
+
+errout:
+	free(path);
+	fprintf(stderr, "Error: cannot join namespace %s\\n", type);
+	return -1;
+
 }
 
 // return 1 if error
 // this function requires root access - todo: fix it!
 int name2pid(const char *name, pid_t *pid) {
 	pid_t parent = getpid();
-	
+
 	DIR *dir;
 	if (!(dir = opendir("/proc"))) {
 		// sleep 2 seconds and try again
@@ -71,7 +72,7 @@ int name2pid(const char *name, pid_t *pid) {
 			exit(1);
 		}
 	}
-	
+
 	struct dirent *entry;
 	char *end;
 	while ((entry = readdir(dir))) {
@@ -90,7 +91,7 @@ int name2pid(const char *name, pid_t *pid) {
 			}
 			free(comm);
 		}
-		
+
 		// look for the sandbox name in /run/firejail/name/<PID>
 		// todo: use RUN_FIREJAIL_NAME_DIR define from src/firejail/firejail.h
 		char *fname;
@@ -187,8 +188,6 @@ char *pid_proc_cmdline(const pid_t pid) {
 	for (i = 0; i < len; i++) {
 		if (buffer[i] == '\0')
 			buffer[i] = ' ';
-//		if (buffer[i] >= 0x80) // execv in progress!!!
-//			return NULL;
 	}
 
 	// return a malloc copy of the command line
@@ -204,6 +203,8 @@ char *pid_proc_cmdline(const pid_t pid) {
 int pid_proc_cmdline_x11_xpra_xephyr(const pid_t pid) {
 	// if comm is not firejail return 0
 	char *comm = pid_proc_comm(pid);
+	if (comm == NULL)
+		return 0;
 	if (strcmp(comm, "firejail") != 0) {
 		free(comm);
 		return 0;
@@ -248,10 +249,10 @@ int pid_proc_cmdline_x11_xpra_xephyr(const pid_t pid) {
 			break;
 		if (strncmp(arg, "--", 2) != 0)
 			break;
-		
+
 		if (strcmp(arg, "--x11=xorg") == 0)
 			return 0;
-		
+
 		// check x11 xpra or xephyr
 		if (strncmp(arg, "--x11", 5) == 0)
 			return 1;
@@ -260,5 +261,24 @@ int pid_proc_cmdline_x11_xpra_xephyr(const pid_t pid) {
 	return 0;
 }
 
+// return 1 if /proc is mounted hidepid, or if /proc/mouns access is denied
+#define BUFLEN 4096
+int pid_hidepid(void) {
+	FILE *fp = fopen("/proc/mounts", "r");
+	if (!fp)
+		return 1;
 
+	char buf[BUFLEN];
+	while (fgets(buf, BUFLEN, fp)) {
+		if (strstr(buf, "proc /proc proc")) {
+			fclose(fp);
+			// check hidepid
+			if (strstr(buf, "hidepid=2") || strstr(buf, "hidepid=1"))
+				return 1;
+			return 0;
+		}
+	}
 
+	fclose(fp);
+	return 0;
+}
